/*
 * schedule.h --
 *
 * This module schedules coroutines based on simulation time.
 * InitSim() must be called before using any routine from this module.
 *
 *	Declarations of ...
 *
 * Copyright 1990 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 * $Header: /usr/local/daedalus/include/RCS/schedule.h,v 1.1 1996/11/06 22:00:31 hari Exp $ SPRITE (Berkeley)
 */

#ifndef _schedule_h_
#define _schedule_h_

#include "mycompat.h"
#include "queue.h"
#include "coproc.h"

#ifdef _XTSIM
typedef unsigned long SimTime;
#else
typedef double SimTime;
#endif _XTSIM

typedef enum { PROCPROC=1, FUNCPROC } ProcType;

#define MAX_USER_QUEUE_NUM 8
typedef struct {
    int         *userQueue[MAX_USER_QUEUE_NUM];
    SimTime     time;
    ProcType    procType;
    int		pid;
} GenProc;

#define NAME_SIZE 80
typedef struct {
    GenProc     genProc;
    COPROCtype  coproc;
    char	name[NAME_SIZE];
} ProcProc;

#define MAX_USER_DATA_NUM 8
typedef struct {
    GenProc     genProc;
    void        (*func)();
    int         *clientData;
    int		userData[MAX_USER_DATA_NUM];
} FuncProc;

extern SimTime	_globalTime;
extern ProcProc	*_spawnTmp;
extern ProcProc	*_curProc;
extern FuncProc	*_curFunc;

/*
 *----------------------------------------------------------------------
 *
 * Spawn --
 *	Spawn is analogous to Unix fork, but duplicates only the current
 *	procedure's *local* variables; therefore, the child process should
 *	never execute a 'return' (call Terminate() instead).
 *	If the child process needs access to the procedure's formal parameters,  *	the formals should be copied to locals before calling Spawn.
 *	Any storage to be shared between processes must be declared global or 
 *	shared via *pointers* which are initialized before calling Spawn.
 *
 * Results:
 *	Spawn returns the process id of the newly created child process to its
 *	parent and zero to the child.
 *
 * Side effects:
 *	Creates and schedules new coroutine thread.
 *
 *----------------------------------------------------------------------
 */

#define Spawn(name) (							\
	SchedProc(0, _spawnTmp = AllocProc(name)),			\
	( CreateCoproc( &(_spawnTmp)->coproc ) ? _spawnTmp : NULL )	\
)

/*
 * Return current simulation time.
 */
#define LocalTime()	(_globalTime)

/*
 * Spend simulation time.
 */
#define Delay(dt)		(Delay1((SimTime) (dt)))
#define SchedProc(dt, proc)	(SchedProc1((SimTime) (dt), (GenProc *) (proc)))
#define ExecuteFunc(p)		(((FuncProc *)(p))->func((p)->clientData));

/*
 * Internal use only.
 */
#define NextProc()	( (GenProc *) QFirst( &ready_q ) )

/*
 * Attach property prop identified by propID to current process.
 * propID may range from 0 to MAX_USER_QUEUE_NUM - 1
 * (To be phased out; use GetCurQueueNode instead.)
 */
#define PutCurProcProp(propID, prop)	\
    (((GenProc *)_curProc)->userQueue[MAX_USER_QUEUE_NUM - 1 - (propID)] = (int *) (prop))

#define PutProcProp(proc, propID, prop)	\
    (((GenProc *)(proc))->userQueue[MAX_USER_QUEUE_NUM - 1 - (propID)] = (int *) (prop))

/*
 * Retreave property identified by propID from process proc.
 * (To be phased out; use GetCurQueueNode instead.)
 */
#define GetProcProp(proc, propID)	\
    (((GenProc *) proc)->userQueue[MAX_USER_QUEUE_NUM - 1 - (propID)])
#define	GetCurProcProp(propID)		\
    (((GenProc *)_curProc)->userQueue[MAX_USER_QUEUE_NUM - 1 - (propID)])

#define GetCurProcQueueNode() (_curProc->userQueue)
#define GetProcQueueNode(proc) (((GenProc *) (proc))->userQueue)

Cextern void		InitSchedule();
Cextern ProcProc	*AllocProc();
Cextern FuncProc	*MakeFunc();
Cextern void		GetFuncArgs();
Cextern void		FreeProc();
Cextern void		FreeFunc();
Cextern void		SchedProc1();
Cextern void		Delay1();
Cextern void		Dispatch();
Cextern void		Terminate();

#endif	/* !_schedule_h_ */
