#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <math.h>

#include "osflags.h"
#include "dbgprintlib.h"
#include "easysocklib.h"
#include "sockinternal.h"

/*************************************************************************
			 Utility  Routines
*************************************************************************/

void Sock_SigPipeHandle()
{
     signal(SIGPIPE, SIG_IGN);
}

static float Sock_Random(void)
{
     float  temp;
#ifdef RANDOM
     while ((temp = (double) (random() % 10000000) / 10000000) == 0.0);
#else
#ifdef RAND
     while ((temp = (double) (rand() % 10000000) / 10000000) == 0.0);
#endif
#endif
     return (temp);
}


/* Return a Poisson variate with specified mean */
static int Sock_PoissonDist(float err_prob)
{
     double   prob;

     prob = -err_prob * log(Sock_Random());
     return (int) (1 / (8.0 * prob));
}

/*
 * Poisson error generator 
 */

void Sock_PoissonError(Socket *sdPtr, char *pktPtr, int size)
{
     static int first_time = 1;
     int num_errors = 0;
     
     while (sdPtr->berState.bytesToNextError <= size) {
	  if (first_time)
	       first_time = 0;
	  else {
	       pktPtr[sdPtr->berState.bytesToNextError] = 
		    pktPtr[sdPtr->berState.bytesToNextError] ^ 
			 (1 << (int)(Sock_Random() * 8));
	       sdPtr->berState.bitsDropped++;
	  }
          sdPtr->berState.bytesToNextError += 
	       Sock_PoissonDist(sdPtr->berState.errorProb[0]);
     }
     sdPtr->berState.bytesToNextError -= size;
     sdPtr->berState.totalBits += size <<3;
}

void Sock_DelayPacket(Socket *sdPtr, int length)
{
     struct timeval curTime, nextTime, time1, time2;
#ifdef NANOSLEEP
     struct timespec sleepTime, left;
#endif
     int usecWait=0;
     static minUsleep = SOCK_MIN_USLEEP;
     int temp;

     nextTime.tv_sec = sdPtr->lastPktTime.tv_sec +
	  sdPtr->lastPktSize/sdPtr->xmitRate;
     nextTime.tv_usec= sdPtr->lastPktTime.tv_usec+
	  ((sdPtr->lastPktSize * 1000000/sdPtr->xmitRate)
	   - ((int)(sdPtr->lastPktSize/sdPtr->xmitRate) * 1000000));
     gettimeofday(&curTime, NULL);
     if (nextTime.tv_sec >= curTime.tv_sec)
	  usecWait = ((nextTime.tv_sec - curTime.tv_sec) * 1000000) +
	       nextTime.tv_usec- curTime.tv_usec;
     sdPtr->lastPktSize = length;
     if (usecWait > 0) {
	  dprintf(120,"nxt sec %d nxt usec %d cur sec %d cur usec %d wait %d\n",
		  nextTime.tv_sec, nextTime.tv_usec, curTime.tv_sec, 
		  curTime.tv_usec, usecWait);
	  if (usecWait > minUsleep) {
	       gettimeofday(&time1, NULL);
#ifdef USLEEP
	       usleep(usecWait);
#else
#ifdef NANOSLEEP
	       sleepTime.tv_sec = nextTime.tv_sec - curTime.tv_sec;
	       sleepTime.tv_nsec = nextTime.tv_usec- curTime.tv_usec * 1000;
	       nanosleep(&sleepTime, &left);
#else
	       sleep(usecWait/1000000);
#endif
#endif
	       gettimeofday(&time2, NULL);
	       temp = (time2.tv_usec - time1.tv_usec) +
		    (time2.tv_sec - time1.tv_sec) * 1000000;
	       if (temp > 1.25 * usecWait) {
		    dprintf(109, "usecWait %d realWait %d minUsleep %d\n", 
			    usecWait, temp, minUsleep);
		    minUsleep = temp;
	       }
	  }
	  sdPtr->lastPktTime.tv_sec = nextTime.tv_sec;
	  sdPtr->lastPktTime.tv_usec = nextTime.tv_usec;
     }
     else {
	  sdPtr->lastPktTime.tv_sec = curTime.tv_sec;
	  sdPtr->lastPktTime.tv_usec = curTime.tv_usec;
     }
}


char *Sock_BroadcastAddr(char *normAddr)
{
     char *newAddr;
     int i,dotCount=0, length;

     length = strlen(normAddr) + 3;
     newAddr = (char *)calloc(length, sizeof(char));
     for (i=0; i< length; i++) {
          if (((normAddr[i] < '0') || (normAddr[i] > '9')) && (normAddr[i] != '.')) {
               free(newAddr);
               return(NULL);
          }
          newAddr[i]=normAddr[i];
          if (normAddr[i] == '.') {
               dotCount++;
               if (normAddr[i+1] == '.') {
                    free(newAddr);
                    return(NULL);
               }
          }
          if (dotCount == 3)
               break;
     }
     newAddr[i+1] = '0';
     newAddr[i+2] = '0';
     newAddr[i+3] = '0';
     dprintf(109,"set broadcast address to %s\n", newAddr);
     return(newAddr);
}

int Sock_IsIPAddr(char *id, int length)
{
     int i,dotCount=0;

     for (i=0; i< length; i++) {
          if (((id[i] < '0') || (id[i] > '9')) && (id[i] != '.'))
               return(0);
          if (id[i] == '.') {
               dotCount++;
               if (id[i+1] == '.')
                    return(0);
          }
     }
     if (dotCount == 3)
          return(1);
     return(0);
}

Sock_Status Sock_NameToSin(char *hostname, Port port, Address *sinPtr) 
{
     int hostAddr, temp;
     char tempstr[1024];
     struct hostent *hp, hpt;

#ifdef INADDR_NONE
     if ((hostAddr = inet_addr(hostname)) != INADDR_NONE) {
#else
     if ((hostAddr = inet_addr(hostname)) != -1) {
#endif
	  dprintf(109, "getting host info from addr\n");
	  memcpy(&(sinPtr->sin_addr), &hostAddr, sizeof(hostAddr));
     }
     else {
	  dprintf(109, "getting host info from name\n");
#ifdef THREADED
	  hp = &hpt;
	  hp = gethostbyname_r(hostname,
			  hp, tempstr, 1024, &temp);
#else
	  hp = gethostbyname(hostname);
#endif
	  endhostent();
	  if (hp == NULL) {
	       sprintf(tempstr, "error setting remote host to %s", hostname);
	       perror(tempstr);
	       endhostent();
	       return SOCK_FAILURE;
	  }
	  memcpy(&(sinPtr->sin_addr), hp->h_addr, hp->h_length);
     }

     sinPtr->sin_family = AF_INET;
     sinPtr->sin_port = htons(port);
     return SOCK_SUCCESS;
}

Sock_Status Sock_SinToName(char **hostnamePtr, Port *port, Address *sinPtr) 
{
     char *returnStr;
     char tempstr[1024];
     struct hostent *hp, hpt;
     int temp;

     *port = ntohs(sinPtr->sin_port);
     if (IN_CLASSD(ntohl(sinPtr->sin_addr.s_addr))) {
	  hp = NULL;
     } else {
#ifdef THREADED
	  hp = &hpt;
	  hp = gethostbyaddr_r((char *)&(sinPtr->sin_addr), 
			       sizeof(struct in_addr), AF_INET,
			       hp, tempstr, 1024, &temp);
#else
	  hp = gethostbyaddr((char *)&(sinPtr->sin_addr), 
			     sizeof(struct in_addr), AF_INET);
#endif
     }
     if ( hp == NULL) {
	  endhostent();
	  dprintf(109,"setting remote to %s\n", inet_ntoa(sinPtr->sin_addr));
	  returnStr = (char *) malloc(strlen(inet_ntoa(sinPtr->sin_addr)));
	  strcpy(returnStr, inet_ntoa(sinPtr->sin_addr));
	  perror("couldn't set remote name by addr");
	  return;
     }
     endhostent();
     dprintf(109, "setting remote to %s\n", hp->h_name);
     returnStr = (char *) malloc(strlen(hp->h_name));
     strcpy(returnStr, hp->h_name);
     *hostnamePtr = returnStr;
}

/*************************************************************************
			 Control Socket Routines
*************************************************************************/

/* recv a ctrl packet on ctrlsdptr and apply operation on sdptr */
Sock_Status Sock_RecvCtrl(Socket *ctrlSdPtr, Socket *sdPtr)
{
     SockCtrlPkt ctrlPkt;
     Sock_Status status=SOCK_SUCCESS;
     int temp,addrLen;
     Address rSin;
     int i;

     dprintf(109, "Sock_RecvCtrl\n");
     addrLen = sizeof(rSin);
     temp = recvfrom(ctrlSdPtr->fd, (char *)&ctrlPkt, sizeof(ctrlPkt), 0,
                     (struct sockaddr *)&(rSin), &addrLen);
     if (temp==-1) {
          dprintf(109,"recvctrl:recv failed\n");
          return(SOCK_FAILURE);
     }

     switch (ntohl(ctrlPkt.reqType)) {

     case SOCK_CTRL_FAREWELL:
          dprintf(109, "recvctrl:farewell ctrl pkt\n");
	  sdPtr->state = SOCK_STATE_CLOSING;
	  break;
     case SOCK_CTRL_HELLO:
          dprintf(109, "recvctrl:hello ctrl pkt\n");
          dprintf(109, "Hello from %s at udp %d tcp %d\n", 
		  ctrlPkt.data.hello.ipAddr, ctrlPkt.data.hello.udpCtrlPort, 
		  ctrlPkt.data.hello.tcpCtrlPort);
	  /* search and add to rhost list XXXX */
	  break;
     case SOCK_CTRL_INIT_HANDOFF:
          /* XXXX should use multicast add of new member */
          dprintf(109, "recvctrl:init handoff ctrl pkt\n");
          /*
             status = Sock_AddDestination(sdPtr, ctrlPkt.data.initHandoff.hostname);
             */
          break;
     case SOCK_CTRL_DO_HANDOFF:
          dprintf(109, "recvctrl:do handoffctrl pkt\n");
          /*
             for (i=0; i<(MAX_MULTICAST_HOSTS); i++) {
             if (sdPtr->activeHosts[i]) {
             if (strcmp(sdPtr->remoteName[i], ctrlPkt.data.initHandoff.hostname))
             status = Sock_DelDestination(sdPtr,
             ctrlPkt.data.initHandoff.hostname);
             }
             }
             */
          break;
     case SOCK_CTRL_SET_RATE:
          /* tested */
          sdPtr->xmitRate = ctrlPkt.data.setRate.rate;
          dprintf(109, "recvctrl:set rate ctrl pkt rate %d\n", sdPtr->xmitRate);
          break;
     case SOCK_CTRL_SET_REMOTE:
          /* tested */
          dprintf(109, "recvctrl:set remote ctrl pkt\n");
          memcpy(&(sdPtr->rSin), &rSin, sizeof(Address));
          Sock_SetRemoteByAddr(sdPtr, &(sdPtr->rSin));
          break;
     case SOCK_CTRL_NEW_CTRL_CHANNEL:
          /* tested 
	     dprintf(109, "recvctrl:new ctrl channel\n");
	     for (i=SOCK_MAX_MULTICAST_HOSTS;
	     i<(SOCK_MAX_MULTICAST_HOSTS+SOCK_MAX_CTRL_HOSTS); i++) {
	     if (!sdPtr->activeHosts[i])
	     break;
	     }
	     sdPtr->ctrlChannel[i] = Sock_Socket(ctrlPkt.data.newCtrlChannel.protocol);
	     status = Sock_Connect(sdPtr->ctrlChannel[i],
	     ctrlPkt.data.newCtrlChannel.hostname,
	     ctrlPkt.data.newCtrlChannel.port);
	     sdPtr->activeHosts[i] = 1;
	     break;
	     */
     default:
          dprintf(109,"unrecognized ctrl pkt: type %d\n",ctrlPkt.reqType);
     }

     return(status);
}

Sock_Status Sock_SendCtrl(Socket *sdPtr, RemoteHost *rhostPtr, 
			  SockCtrlPkt *ctrlPktPtr)
{
     Sock_Status status=SOCK_FAILURE;
     
     
     if (rhostPtr->tcpCtrl != NULL)
	  Sock_Send(rhostPtr->tcpCtrl, (char *) ctrlPktPtr, sizeof(*ctrlPktPtr));
     else {
	  if (sendto(sdPtr->udpCtrlChannel->fd, (char *)ctrlPktPtr, 
		     sizeof(*ctrlPktPtr), 0,
		     (struct sockaddr *)&(rhostPtr->udpRSin),
		     sizeof(rhostPtr->udpRSin)) == -1) {
	       perror("Sock_SendCtrl: error in sendto");
	       return(SOCK_FAILURE);
	  }
     }
     return(SOCK_SUCCESS);
}

Sock_Status Sock_CtrlSendFarewell(Socket *sdPtr, RemoteHost *rhostPtr)
{
     SockCtrlPkt ctrlPkt;
     
     dprintf(109, "sendctrl:farewell ctrl pkt\n");
     strcpy(ctrlPkt.data.farewell.ipAddr, sdPtr->localIPAddr);
     ctrlPkt.reqType = htonl(SOCK_CTRL_FAREWELL);
     return(Sock_SendCtrl(sdPtr, rhostPtr, &ctrlPkt));
}
/*
   Sock_Status Sock_CtrlSetAddr(Socket *ctrlSdPtr, Port port)
   {
	SockCtrlPkt ctrlPkt;

	ctrlPkt.data.setRemote.port = port;
	ctrlPkt.reqType = SOCK_CTRL_SET_REMOTE;
	return(Sock_SendCtrl(ctrlSdPtr, &ctrlPkt));
   }
   
   Sock_Status Sock_CtrlSetCtrlAddr(Socket *ctrlSdPtr, Port port)
   {
	SockCtrlPkt ctrlPkt;

	ctrlPkt.data.setRemote.port = port;
	ctrlPkt.reqType = SOCK_CTRL_SET_REMOTE_CTRL;
	return(Sock_SendCtrl(ctrlSdPtr, &ctrlPkt));
   }
   
   Socket *Sock_CtrlNewCtrl(Socket *ctrlSdPtr, char *hostname,
			    Port port, int protocol)
			    {
				 SockCtrlPkt ctrlPkt;
				 Socket *sdPtr, *cdPtr;
				 Sock_Status status;

				 sdPtr = Sock_Socket(protocol);
				 status = Sock_Listen(sdPtr, port);
				 ctrlPkt.data.newCtrlChannel.port = sdPtr->localPort;
				 ctrlPkt.data.newCtrlChannel.protocol = protocol;
				 strcpy(ctrlPkt.data.newCtrlChannel.hostname, sdPtr->localName);
				 ctrlPkt.reqType = SOCK_CTRL_NEW_CTRL_CHANNEL;
				 status = Sock_SendCtrl(ctrlSdPtr, &ctrlPkt);
				 cdPtr = Sock_Accept(sdPtr);
				 Sock_Close(sdPtr);
				 return(cdPtr);
			    }
			    
			    Sock_Status Sock_CtrlSetRate(Socket *ctrlSdPtr, int rate)
			    {
				 SockCtrlPkt ctrlPkt;

				 ctrlPkt.data.setRate.rate= rate;
				 ctrlPkt.reqType = SOCK_CTRL_SET_RATE;
				 return(Sock_SendCtrl(ctrlSdPtr, &ctrlPkt));
			    }
			    
			    Sock_Status Sock_CtrlInitHandoff(Socket *ctrlSdPtr, char *newHost)
			    {
				 SockCtrlPkt ctrlPkt;

				 dprintf(109, "entering init handoff to %s\n", newHost);
				 strcpy(ctrlPkt.data.initHandoff.hostname, newHost);
				 ctrlPkt.reqType = SOCK_CTRL_INIT_HANDOFF;
				 return(Sock_SendCtrl(ctrlSdPtr, &ctrlPkt));
			    }
			    
			    Sock_Status Sock_CtrlDoHandoff(Socket *ctrlSdPtr, char *newHost)
			    {
				 SockCtrlPkt ctrlPkt;

				 dprintf(109, "entering init handoff to %s\n", newHost);
				 strcpy(ctrlPkt.data.doHandoff.hostname, newHost);
				 ctrlPkt.reqType = SOCK_CTRL_DO_HANDOFF;
				 return(Sock_SendCtrl(ctrlSdPtr, &ctrlPkt));
			    }
			    /*
			       
/*************************************************************************
  				Socket Routines
*************************************************************************/

Socket *Sock_Socket(PROTOCOLS protocol)
{
     Socket *sdPtr;
     struct protoent *proto;
     int i, j;
     char *protoStr;
     Sock_Status temp;
     Address tempAddr;

     dprintf(112, "Sock_Socket\n");
     j = sizeof(i);
     sdPtr = (Socket *)malloc(sizeof(Socket));
     memset(sdPtr, 0, sizeof(Socket));
     sdPtr->isCtrl = SOCK_NOT_CTRL;
     switch (protocol) {
     case SOCK_TCP_CTRL:
          sdPtr->isCtrl = SOCK_CTRL;
     case SOCK_TCP:
          sdPtr->protocol = SOCK_TCP;
          sdPtr->type = SOCK_STREAM;
          protoStr="tcp";
          break;
     case SOCK_UDP_CTRL:
          sdPtr->isCtrl = SOCK_CTRL;
     case SOCK_UDP:
          sdPtr->protocol = SOCK_UDP;
          sdPtr->type = SOCK_DGRAM;
          protoStr="udp";
          break;
     case SOCK_UDP_MULTI_CTRL:
          sdPtr->isCtrl = SOCK_CTRL;
     case SOCK_UDP_MULTI:
          sdPtr->protocol = SOCK_UDP_MULTI;
          sdPtr->type = SOCK_DGRAM;
          protoStr="udp";
          break;
     case SOCK_IP:
          sdPtr->protocol = SOCK_IP;
          sdPtr->type = SOCK_RAW;
          protoStr="ip";
          break;
     default:
          perror("Sock_socket: invalid protocol");
          free(sdPtr);
          return((Socket *)SOCK_FAILURE);
     }
     proto = getprotobyname(protoStr);
     endprotoent();

     /* get the socket */
     if ((sdPtr->fd = socket(AF_INET, sdPtr->type, proto->p_proto)) == -1) {
          perror("Sock_socket: error opening socket");
          return((Socket *)SOCK_FAILURE);
     }

     /* set defaults */
     i=1;
     if (sdPtr->protocol==SOCK_TCP) {
	  if (setsockopt(sdPtr->fd,proto->p_proto, 
			 TCP_NODELAY,(char *)&i,j) < 0)
	       perror("Sock_Socket: tcp_nodelay");
     }
     if (setsockopt(sdPtr->fd, SOL_SOCKET, SO_BROADCAST, (char *)&i,j) < 0)
          perror("Sock_Socket: so_broadcast");
     if (setsockopt(sdPtr->fd, SOL_SOCKET, SO_REUSEADDR, (char *)&i,j) < 0)
          perror("Sock_Socket: reuseaddr");
     sdPtr->remotePort = 0;
     Sock_SetBufferSize(sdPtr, SOCK_DEFAULT_BUF_SIZE);
     gethostname(sdPtr->localName, SOCK_MAX_HOST_NAME_LEN);
     Sock_NameToSin(sdPtr->localName, 10000, &tempAddr);
     strcpy(sdPtr->localIPAddr, inet_ntoa(tempAddr.sin_addr));
     Sock_SetTimeout(sdPtr, SOCK_NO_TIMEOUT);
     gettimeofday(&(sdPtr->lastPktTime), NULL);
     sdPtr->ttl = SOCK_DEFAULT_TTL;
     sdPtr->xmitRate = 0;
     sdPtr->localPort = 0;
     sdPtr->block = SOCK_BLOCK;
     sdPtr->broadcast = SOCK_NOT_BROADCAST;
     sdPtr->compatible = SOCK_NOT_COMPATIBLE;
     sdPtr->mobile = SOCK_NOT_MOBILE;
     sdPtr->udpCtrlChannel = NULL;
     sdPtr->tcpCtrlChannel = NULL;
     sdPtr->numActiveRHost = 0;
     sdPtr->rHost = NULL;
     sdPtr->lFlags = 0;

     sdPtr->berState.ErrorGen = NULL;
     sdPtr->state = SOCK_STATE_ALLOCATED;
     return(sdPtr);
}

Sock_Status Sock_Connect(Socket *sdPtr, char *hostname, Port port)
{
     int status;
     char tempstr[128], c;
     Port remoteCtrlPort;
     int length;
     struct sockaddr_in local;
     
     dprintf(112, "Sock_Connect %s %d\n", hostname, port);
     
     sdPtr->rHost = (RemoteHost *) malloc(sizeof(RemoteHost));
     memset(sdPtr->rHost, 0, sizeof(RemoteHost));
     if (Sock_SetRemoteByName(sdPtr, hostname, port) == SOCK_FAILURE)
          return(SOCK_FAILURE);
     
     if (sdPtr->compatible == SOCK_NOT_COMPATIBLE && 
	 sdPtr->isCtrl != SOCK_CTRL) {
	  sdPtr->tcpCtrlChannel = Sock_Socket(SOCK_TCP_CTRL);
	  Sock_Listen(sdPtr->tcpCtrlChannel, SOCK_ANY_PORT);
	  sdPtr->udpCtrlChannel = Sock_Socket(SOCK_UDP_CTRL);
	  Sock_Listen(sdPtr->udpCtrlChannel, SOCK_ANY_PORT);
     }
     
     /* create connection to remote host */
     switch (sdPtr->protocol) {
     case SOCK_UDP:
	  if (sdPtr->state == SOCK_STATE_ALLOCATED) {
	       sdPtr->lSin.sin_family = AF_INET;
	       sdPtr->lSin.sin_port = htons(0);
	       sdPtr->lSin.sin_addr.s_addr = htonl(INADDR_ANY);
	       if ((bind(sdPtr->fd,
			 (struct sockaddr *)&sdPtr->lSin,
			 sizeof(sdPtr->lSin))) == -1) {
		    perror("Sock_connect: error binding socket");
		    return(SOCK_FAILURE);
	       }
	  }
          if (sdPtr->compatible == SOCK_NOT_COMPATIBLE && 
	      sdPtr->isCtrl != SOCK_CTRL) {
	       SockCtrlPkt temp;
               /* send "Hello connect me" packet */
	       strcpy(temp.data.hello.ipAddr, sdPtr->localIPAddr);
	       temp.data.hello.tcpCtrlPort = htons(Sock_GetTCPCtrlPort(sdPtr));
	       temp.data.hello.udpCtrlPort = htons(Sock_GetUDPCtrlPort(sdPtr));
	       Sock_Send(sdPtr, (char *)&temp, sizeof(temp));
	       /* receive "Hello accepted" packet */
	       Sock_RecvFull(sdPtr, (char *)&temp, sizeof(temp));
	       sdPtr->rHost->udpCtrlPort=ntohs(temp.data.hello.udpCtrlPort);
	       sdPtr->rHost->tcpCtrlPort=ntohs(temp.data.hello.tcpCtrlPort);
	       sdPtr->rHost->tcpCtrl=NULL;
	       sdPtr->rHost->next=NULL;
	       strcpy(sdPtr->rHost->remoteIPAddr, temp.data.hello.ipAddr);
	       Sock_NameToSin(sdPtr->rHost->remoteIPAddr, 
			      sdPtr->rHost->udpCtrlPort,
			      &(sdPtr->rHost->udpRSin));
	  }
	  sdPtr->numActiveRHost=1;
          break;
#ifdef MULTICAST
      case SOCK_UDP_MULTI:
          sdPtr->lSin = sdPtr->rSin;
	  if (IN_CLASSD(ntohl(sdPtr->rSin.sin_addr.s_addr))) {
	       struct ip_mreq mr;

	       if ((bind(sdPtr->fd,
			 (struct sockaddr *)&sdPtr->lSin,
			 sizeof(sdPtr->lSin))) == -1) {
		    sdPtr->lSin.sin_addr.s_addr = htonl(INADDR_ANY);
		    if ((bind(sdPtr->fd,
			      (struct sockaddr *)&sdPtr->lSin,
			      sizeof(sdPtr->lSin))) == -1) {
			 perror("Sock_connect: error binding socket");
			 return(SOCK_FAILURE);
		    }
	       }
	       
	       mr.imr_multiaddr.s_addr = sdPtr->rSin.sin_addr.s_addr;
	       mr.imr_interface.s_addr = INADDR_ANY;
	       if (setsockopt(sdPtr->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, 
			      (char *)&mr, sizeof(mr)) < 0) {
		    perror("connect: IP_ADD_MEMBERSHIP");
		    return(SOCK_FAILURE);
	       }
	  } else 
	       return(SOCK_FAILURE);
	  c=0;
          if (setsockopt(sdPtr->fd, IPPROTO_IP, IP_MULTICAST_LOOP, &c, 1) < 0)
	       perror("Sock_Connect: ip_multicast_loop");
          if (setsockopt(sdPtr->fd, IPPROTO_IP, IP_MULTICAST_TTL,&sdPtr->ttl,1)
              < 0)
	       perror("Sock_Connect: ip_multicast_ttl");
          if (sdPtr->compatible == SOCK_NOT_COMPATIBLE && 
	      sdPtr->isCtrl != SOCK_CTRL) {
	       SockCtrlPkt temp;
	       sdPtr->rHost->udpCtrlPort=port+1;
	       sdPtr->rHost->tcpCtrlPort=-1;
	       sdPtr->rHost->tcpCtrl=NULL;
	       sdPtr->rHost->next=NULL;
	       sdPtr->rHost->remoteName="MULTICAST";
	       Sock_NameToSin(hostname, sdPtr->rHost->udpCtrlPort, 
			 &(sdPtr->rHost->udpRSin));
               /* send "Hello" packet */
	       strcpy(temp.data.hello.ipAddr, sdPtr->localIPAddr);
	       temp.data.hello.tcpCtrlPort = htons(Sock_GetTCPCtrlPort(sdPtr));
	       temp.data.hello.udpCtrlPort = htons(Sock_GetUDPCtrlPort(sdPtr));
	       Sock_SendCtrl(sdPtr, sdPtr->rHost, &temp);
	       /* no welcome/accepted pkt */
	  }
	  sdPtr->numActiveRHost=1;
	  break;
#endif
     default:
	  if ((status = connect(sdPtr->fd, (struct sockaddr *)&sdPtr->rSin,
				sizeof(sdPtr->rSin))) == -1) {
	       perror("Sock_connect: Error connecting");
	       return(SOCK_FAILURE);
	  }
	  if (sdPtr->compatible == SOCK_NOT_COMPATIBLE && 
	      sdPtr->isCtrl != SOCK_CTRL) {
	       SockCtrlPkt temp;
	       /* send "Hello connect" packet */
	       strcpy(temp.data.hello.ipAddr, sdPtr->localIPAddr);
	       temp.data.hello.tcpCtrlPort = htons(Sock_GetTCPCtrlPort(sdPtr));
	       temp.data.hello.udpCtrlPort = htons(Sock_GetUDPCtrlPort(sdPtr));
	       Sock_Send(sdPtr, (char *)&temp, sizeof(temp));
	       /* receive "Hello accept" packet */
	       Sock_RecvFull(sdPtr, (char *)&temp, sizeof(temp));
	       sdPtr->rHost->udpCtrlPort=ntohs(temp.data.hello.udpCtrlPort);
	       sdPtr->rHost->tcpCtrlPort=ntohs(temp.data.hello.tcpCtrlPort);
	       sdPtr->rHost->tcpCtrl=NULL;
	       sdPtr->rHost->next=NULL;
	       strcpy(sdPtr->rHost->remoteIPAddr, temp.data.hello.ipAddr);
	       Sock_NameToSin(sdPtr->rHost->remoteIPAddr, 
			      sdPtr->rHost->udpCtrlPort,
			      &(sdPtr->rHost->udpRSin));
	  }
	  sdPtr->numActiveRHost=1;
     }  
     sdPtr->side = SOCK_SIDE_CLIENT;
     length = sizeof(Address);
     if ((status = getsockname(sdPtr->fd, (struct sockaddr *)&sdPtr->lSin,
			       &length)) == -1) {
	  perror("Sock_Connect: error getting socket name");
	  return(SOCK_FAILURE);
     }
     sdPtr->localPort = ntohs(sdPtr->lSin.sin_port);
     dprintf(109,"connected from inet port %d to %d tcp, %d udp\n",
	     sdPtr->localPort, sdPtr->rHost->tcpCtrlPort, 
	     sdPtr->rHost->udpCtrlPort );
     sdPtr->state = SOCK_STATE_CONNECTED;
     return(SOCK_SUCCESS);
}
     
Sock_Status Sock_Listen(Socket *sdPtr, Port port)
{
     int temp;
     int length;
     struct sockaddr_in *local;

     dprintf(112, "Sock_Listen %d \n",port);
     sdPtr->lSin.sin_family = AF_INET;
     sdPtr->lSin.sin_port = htons(port);
     sdPtr->lSin.sin_addr.s_addr = htonl(INADDR_ANY);
     if ((bind(sdPtr->fd,
	       (struct sockaddr *)&sdPtr->lSin,
	       sizeof(sdPtr->lSin))) == -1) {
	  perror("Sock_listen: error binding socket");
	  return(SOCK_FAILURE);
     }
     length = sizeof(Address);
     if ((temp = getsockname(sdPtr->fd, (struct sockaddr *)&sdPtr->lSin,
			     &length)) == -1) {
	  perror("Sock_Listen: error getting socket name");
	  return(SOCK_FAILURE);
     }
     sdPtr->localPort = ntohs(sdPtr->lSin.sin_port);
     Sock_NameToSin(sdPtr->localName, sdPtr->localPort, &sdPtr->lSin);
     dprintf(109,"listening at inet port %d\n",sdPtr->localPort);
     if (sdPtr->protocol == SOCK_TCP) {
	  if ((temp = listen(sdPtr->fd, 5)) == -1) {
	       perror("Sock_listen: error listening");
	       return(SOCK_FAILURE);
	  }
     }
     sdPtr->state = SOCK_STATE_LISTENING;
     return(SOCK_SUCCESS);
}

Socket *Sock_Accept(Socket *sdPtr)
{
     int ts;
     Socket *cdPtr, *tdPtr;
     int addrLen;
     Port port,remotePort;

     dprintf(112, "Sock_Accept\n");
     cdPtr = (Socket *)malloc(sizeof(Socket));
     memcpy(cdPtr, sdPtr, sizeof(Socket));
     cdPtr->rHost = (RemoteHost *) malloc(sizeof(RemoteHost));
     memset(cdPtr->rHost, 0, sizeof(RemoteHost));

     if (cdPtr->compatible == SOCK_NOT_COMPATIBLE && 
	 cdPtr->isCtrl != SOCK_CTRL) {
	  cdPtr->tcpCtrlChannel = Sock_Socket(SOCK_TCP_CTRL);
	  cdPtr->udpCtrlChannel = Sock_Socket(SOCK_UDP_CTRL);
	  Sock_Listen(cdPtr->udpCtrlChannel, SOCK_ANY_PORT);
	  Sock_Listen(cdPtr->tcpCtrlChannel, SOCK_ANY_PORT);
     }
     switch (cdPtr->protocol) {
     case SOCK_IP:
	  return(cdPtr);
     case SOCK_UDP:
	  if (cdPtr->compatible == SOCK_NOT_COMPATIBLE) {
	       SockCtrlPkt temp;
	       /* expect a "connect me" packet */
	       Sock_RecvFull(cdPtr, (char *)&temp, sizeof(temp));
	       cdPtr->rHost->udpCtrlPort=ntohs(temp.data.hello.udpCtrlPort);
	       cdPtr->rHost->tcpCtrlPort=ntohs(temp.data.hello.tcpCtrlPort);
	       cdPtr->rHost->tcpCtrl=NULL;
	       cdPtr->rHost->next=NULL;
	       strcpy(cdPtr->rHost->remoteIPAddr, temp.data.hello.ipAddr);
	       Sock_SetRemoteByAddr(cdPtr, &(cdPtr->lastRSin));
	       Sock_NameToSin(cdPtr->rHost->remoteIPAddr, 
			      cdPtr->rHost->udpCtrlPort,
			      &(cdPtr->rHost->udpRSin));
	       /* send and "accepted" packet */
	       strcpy(temp.data.hello.ipAddr, cdPtr->localIPAddr);
	       temp.data.hello.tcpCtrlPort = htons(Sock_GetTCPCtrlPort(cdPtr));
	       temp.data.hello.udpCtrlPort = htons(Sock_GetUDPCtrlPort(cdPtr));
	       Sock_Send(cdPtr, (char *)&temp, sizeof(temp));
	       sdPtr->fd = -1;
	  }
	  cdPtr->numActiveRHost=1;
	  break;
     case SOCK_TCP:
	  addrLen = sizeof(cdPtr->rSin);
	  if ((ts = accept(cdPtr->fd, (struct sockaddr *)&(cdPtr->rSin),
			   &addrLen))==-1) {
	       perror("Sock_accept: error accepting connection");
	       return((Socket *)SOCK_FAILURE);
	  }
	  cdPtr->fd = ts;
	  cdPtr->fd = ts;
	  if (cdPtr->compatible == SOCK_NOT_COMPATIBLE) {
	       SockCtrlPkt temp;
	       /* expect a "connect me" packet */
	       Sock_RecvFull(cdPtr, (char *)&temp, sizeof(temp));
	       cdPtr->rHost->udpCtrlPort=ntohs(temp.data.hello.udpCtrlPort);
	       cdPtr->rHost->tcpCtrlPort=ntohs(temp.data.hello.tcpCtrlPort);
	       cdPtr->rHost->tcpCtrl=NULL;
	       cdPtr->rHost->next=NULL;
	       strcpy(cdPtr->rHost->remoteIPAddr, temp.data.hello.ipAddr);
	       Sock_NameToSin(cdPtr->rHost->remoteIPAddr,
			      cdPtr->rHost->udpCtrlPort,
			      &(cdPtr->rHost->udpRSin));
	       /* send and "accepted" packet */
	       strcpy(temp.data.hello.ipAddr, cdPtr->localIPAddr);
	       temp.data.hello.tcpCtrlPort = htons(Sock_GetTCPCtrlPort(cdPtr));
	       temp.data.hello.udpCtrlPort = htons(Sock_GetUDPCtrlPort(cdPtr));
	       Sock_Send(cdPtr, (char *)&temp, sizeof(temp));
	  }
	  Sock_SetRemoteByAddr(cdPtr, &(cdPtr->rSin));
	  cdPtr->numActiveRHost=1;
	  break;
#ifdef MULTICAST
     case SOCK_UDP_MULTI:
	  dprintf(109, "Sock_Accept: multicast\n");
	  break;
#endif
     default:
	  dprintf(109, "Sock_Accept: other protocol\n");
     }

     dprintf(109,"connected from inet port %d to %d tcp, %d udp\n",
	     cdPtr->localPort, cdPtr->rHost->tcpCtrlPort, 
	     cdPtr->rHost->udpCtrlPort );
     cdPtr->side = SOCK_SIDE_SERVER;
     cdPtr->state = SOCK_STATE_ACCEPTED;
     return(cdPtr);
}

Sock_Status Sock_Close(Socket *sdPtr)
{
     int temp=SOCK_SUCCESS,i;
     RemoteHost *rhPtr;
     
     dprintf(112, "Sock_Close\n");
     if (sdPtr->state == SOCK_STATE_CLOSED)
	  return(SOCK_SUCCESS);

     /* send "farewell" message to first rhost */

     while (sdPtr->numActiveRHost > 0) {
	  rhPtr=sdPtr->rHost;
	  if (rhPtr == NULL)
	       dprintf(101,"Sock_close1: # of rhost error\n");
	  if (sdPtr->compatible == SOCK_NOT_COMPATIBLE) 
	       Sock_CtrlSendFarewell(sdPtr, rhPtr);
	  sdPtr->rHost = rhPtr->next;
	  sdPtr->numActiveRHost--;
	  if (rhPtr->tcpCtrl != NULL) {
	       temp = Sock_Close(rhPtr->tcpCtrl);
	       temp = Sock_Free(sdPtr->rHost[i].tcpCtrl);
	  }
     }

     
     if (sdPtr->rHost != NULL)
	  dprintf(101,"Sock_close: # of rhost error\n");
     if (sdPtr->tcpCtrlChannel != NULL) {
	  Sock_Close(sdPtr->tcpCtrlChannel);
	  Sock_Free(sdPtr->tcpCtrlChannel);
     }
     if (sdPtr->udpCtrlChannel != NULL) {
	  Sock_Close(sdPtr->udpCtrlChannel);
	  Sock_Free(sdPtr->udpCtrlChannel);
     }

     if ((sdPtr)->fd != -1) {
	  temp = shutdown(sdPtr->fd, 2);
	  close(sdPtr->fd);
     }

     sdPtr->state = SOCK_STATE_CLOSED;
     return((Sock_Status)temp);
}

Sock_Status Sock_Free(Socket *sdPtr)
{
     int temp,i;

     if (sdPtr->state != SOCK_STATE_CLOSED)
	  return(SOCK_FAILURE);

     free(sdPtr);
     return(SOCK_SUCCESS);
}

/*************************************************************************
  send/recv routines
  return val: SOCK_FAILURE or bytes xmitd/recvd
  *************************************************************************/

void Check_Ctrl(Socket *sdPtr)
{
     RemoteHost *rhPtr;
     fd_set fdvar;
     static struct timeval poll = {0,0};
     int temp;

     rhPtr = sdPtr->rHost;
     FD_ZERO(&fdvar);
     while (rhPtr != NULL) {
	  if (rhPtr->tcpCtrl != NULL)
	       FD_SET(rhPtr->tcpCtrl->fd, &fdvar);
	  rhPtr = rhPtr->next;
     }
     FD_SET(sdPtr->udpCtrlChannel->fd, &fdvar);
     FD_SET(sdPtr->tcpCtrlChannel->fd, &fdvar);
     if ((temp = select(FD_SETSIZE, &fdvar, (fd_set *)0, (fd_set *)0,
			&poll)) <= 0) {
	  dprintf(120,"Sock_Recv: no ctrl info\n");
     }
     else {
	  dprintf(112,"Sock_Recv: ctrl info temp = %d\n",temp);
	  if (FD_ISSET(sdPtr->udpCtrlChannel->fd, &fdvar))
	       Sock_RecvCtrl(sdPtr->udpCtrlChannel,sdPtr);
	  if (FD_ISSET(sdPtr->tcpCtrlChannel->fd, &fdvar))
	       /* XXXX */
	       dprintf(120,"Sock_Recv: accept new tcp ctrl\n");
	  rhPtr = sdPtr->rHost;
	  while (rhPtr != NULL) {
	       if (rhPtr->tcpCtrl != NULL)
		    if (FD_ISSET(rhPtr->tcpCtrl->fd, &fdvar))
			 Sock_RecvCtrl(rhPtr->tcpCtrl,sdPtr); 
	       rhPtr = rhPtr->next;
	  }
     }
     
}

Sock_Status Sock_Recv(Socket *sdPtr, char *buffer, int length)
{
     Sock_Status status;
     fd_set fdvar;
     int temp,i;
     static struct timeval defltTout = {SOCK_CTRL_TIMEOUT_SEC, SOCK_CTRL_TIMEOUT_USEC};

     dprintf(112, "Sock_Recv %d\n", length);
     while (1) {
	  if (sdPtr->useCtrl == SOCK_USE_CTRL)
	       Check_Ctrl(sdPtr);

	  if (sdPtr->state == SOCK_STATE_CLOSED || 
	      sdPtr->state == SOCK_STATE_CLOSING)
	       return(SOCK_FD_CLOSED);

	  FD_ZERO(&fdvar);
	  FD_SET(sdPtr->fd, &fdvar);

	  if (sdPtr->block == SOCK_NON_BLOCK)
	       break;

	  if (sdPtr->toutPtr == NULL) {
	       if ((temp = select(FD_SETSIZE, &fdvar, (fd_set *)0, (fd_set *)0,
				  &defltTout)) > 0) {
		    break;
	       }
	  }
	  else {
	       if ((temp = select(FD_SETSIZE, &fdvar, (fd_set *)0, (fd_set *)0,
				  sdPtr->toutPtr)) == 0) {
		    dprintf(102,"Sock_Recv: timeout\n");
		    return(SOCK_TIMEOUT);
	       }
	  }
     }

     switch (sdPtr->type) {
     case SOCK_STREAM:
	  status = Sock_RecvStream(sdPtr, buffer, length);
	  break;
     case SOCK_DGRAM:
	  status = Sock_RecvDgram(sdPtr, buffer, length);
	  break;
     default:
	  dprintf(101, "unknown socket type\n");
	  status = SOCK_FAILURE;
     }

     return(status);
}

Sock_Status Sock_RecvStream(Socket *sdPtr, char *buffer, int length)
{
     int temp;
     fd_set fdvar;

     temp = read(sdPtr->fd, buffer, length);
     if(temp == 0) {
	  if (errno = EACCES)
	       return(SOCK_FD_CLOSED);
	  return(SOCK_FAILURE);
     }
     if (temp == -1)
	  dprintf(101,"Sock_recv: error in recv\n");
     if (temp != length)
	  dprintf(109, "Sock_Recv: warning short read %d read %d wanted\n",
		  temp, length);
     return(temp);
}

Sock_Status Sock_RecvDgram(Socket *sdPtr, char *buffer, int length)
{
     int addrLen, temp;
     fd_set fdvar;

     addrLen = sizeof(sdPtr->lastRSin);
     temp = recvfrom(sdPtr->fd, buffer, length, 0,
		     (struct sockaddr *)&(sdPtr->lastRSin),
		     &addrLen);
     if (temp == -1) {
	  dprintf(101,"Sock_RecvDgram: bad recvfrom\n");
	  return(SOCK_FAILURE);
     }
     if (temp != length)
	  dprintf(109, "Sock_RecvDgram: warning short read %d read %d wanted\n",
		  temp, length);
     return(temp);
}


Sock_Status Sock_RecvFull(Socket *sdPtr, char *buffer, int length)
{
     Sock_Status status;
     fd_set fdvar;
     int temp,i;
     static struct timeval defltTout = {SOCK_CTRL_TIMEOUT_SEC, SOCK_CTRL_TIMEOUT_USEC};

     dprintf(112, "Sock_RecvFull %d\n", length);

     while (1) {
	  if (sdPtr->useCtrl == SOCK_USE_CTRL)
	       Check_Ctrl(sdPtr);

	  if (sdPtr->state == SOCK_STATE_CLOSED || 
	      sdPtr->state == SOCK_STATE_CLOSING)
	       return(SOCK_FD_CLOSED);

	  FD_ZERO(&fdvar);
	  FD_SET(sdPtr->fd, &fdvar);

	  if (sdPtr->toutPtr == NULL) {
	       if ((temp = select(FD_SETSIZE, &fdvar, (fd_set *)0, (fd_set *)0,
				  &defltTout)) > 0) {
		    break;
	       }
	  }
	  else {
	       if ((temp = select(FD_SETSIZE, &fdvar, (fd_set *)0, (fd_set *)0,
				  sdPtr->toutPtr)) == 0) {
		    dprintf(102,"Sock_Recv: timeout\n");
		    return(SOCK_TIMEOUT);
	       }
	  }
     }

     switch (sdPtr->type) {
     case SOCK_STREAM:
	  status = Sock_RecvStreamFull(sdPtr, buffer, length);
	  break;
     case SOCK_DGRAM:
	  status = Sock_RecvDgramFull(sdPtr, buffer, length);
	  break;
     default:
	  dprintf(101, "unknown socket type\n");
	  status = SOCK_FAILURE;
     }

     return(status);
}


/*
 *                      M R E A D
 *
 * This function performs the function of a read(II) but will
 * call read(II) multiple times in order to get the requested
 * number of characters.  This can be necessary because
 * network connections don't deliver data with the same
 * grouping as it is written with.  Written by Robert S. Miles, BRL.
 */
Sock_Status Sock_RecvStreamFull(Socket *sdPtr, register char *buffer, int length)
{
     register unsigned     count = 0;
     register int          nread;
     fd_set fdvar;

     FD_ZERO(&fdvar);
     FD_SET(sdPtr->fd, &fdvar);

     do {
	  if (select(FD_SETSIZE, &fdvar, (fd_set *)0, (fd_set *)0,
		     sdPtr->toutPtr) == 0) {
	       dprintf(102,"Sock_Recv: timeout\n");
	       return(SOCK_TIMEOUT);
	  }
	  nread = read(sdPtr->fd, buffer, length-count);
	  if(nread < 0)  {
	       dprintf(101,"Sock_RecvFull: error in read\n");
	       return(SOCK_FAILURE);
	  }
	  if(nread == 0) {
	       if (errno = EACCES)
		    return(SOCK_FD_CLOSED);
	       return(SOCK_FAILURE);
	  }
	  count += (unsigned)nread;
	  buffer += nread;
     } while(count < length);

     return((int)count);
}


Sock_Status Sock_RecvDgramFull(Socket *sdPtr, char *buffer, int length)
{
     fd_set fdvar;
     int addrLen;
     register unsigned     count = 0;
     register int          nread;

     FD_ZERO(&fdvar);
     FD_SET(sdPtr->fd, &fdvar);

     do {
	  if (select(FD_SETSIZE, &fdvar, (fd_set *)0, (fd_set *)0,
		     sdPtr->toutPtr) == 0) {
	       dprintf(102,"Sock_RecvDgramFull: timeout\n");
	       return(SOCK_TIMEOUT);
	  }
	  addrLen = sizeof(sdPtr->lastRSin);
	  nread=recvfrom(sdPtr->fd,buffer,length,0,
			 (struct sockaddr *)&(sdPtr->lastRSin),
			 &addrLen);
	  if (nread == -1) {
	       perror("Sock_RecvDgramFull: bad recvfrom");
	       return(SOCK_FAILURE);
	  }
	  if (nread == 0)
	       return((int)count);
	  count += (unsigned)nread;
	  buffer += nread;
     } while(count < length);

     return((int)count);
}

Sock_Status Sock_Send(Socket *sdPtr, char *buffer, int length)
{
     static struct timeval poll = {0,0};
     Sock_Status status;
     fd_set fdvar;

     dprintf(112, "Sock_Send %d\n",length);
     if (sdPtr->useCtrl == SOCK_USE_CTRL)
	  Check_Ctrl(sdPtr);

     if (sdPtr->state == SOCK_STATE_CLOSED ||
	 sdPtr->state == SOCK_STATE_CLOSING)
               return(SOCK_FD_CLOSED);

     if (sdPtr->xmitRate > 0)
	  Sock_DelayPacket(sdPtr, length);

     if (sdPtr->berState.ErrorGen != NULL) 
	  (sdPtr->berState.ErrorGen)(sdPtr, buffer, length);

     switch (sdPtr->type) {
     case SOCK_STREAM:
	  status = Sock_SendStream(sdPtr, buffer, length);
	  break;
     case SOCK_DGRAM:
	  status = Sock_SendDgram(sdPtr, buffer, length);
	  break;
     default:
	  dprintf(101, "unknown socket type\n");
	  status = SOCK_FAILURE;
     }
     return(status);
}

Sock_Status Sock_SendStream(Socket *sdPtr, char *buffer, int length)
{
     int temp;
     if ((temp = send(sdPtr->fd, buffer, length, sdPtr->lFlags)) == -1) {
	  perror("Sock_send: error in send");
	  if (errno == EPIPE) {
	       sdPtr->state == SOCK_STATE_CLOSED;
	       return(SOCK_FD_CLOSED);
	  }
	  return(SOCK_FAILURE);
     }
     return(temp);
}

/* send length bytes from buffer to fsin (faster) */
Sock_Status Sock_SendDgram(Socket *sdPtr, char *buffer, int length)
{
     int temp;
     fd_set fdvar;

     FD_ZERO(&fdvar);
     FD_SET(sdPtr->fd, &fdvar);
     select(FD_SETSIZE, (fd_set *)0,  &fdvar, (fd_set *)0, NULL);
     while (1) {
	  if ((temp = sendto(sdPtr->fd, buffer, length, sdPtr->lFlags,
			     (struct sockaddr *)&(sdPtr->rSin),
			     sizeof(sdPtr->rSin))) == -1) {
	       if (errno == ENOBUFS)
		    continue;
	       perror("Sock_SendDgram: error in sendto");
	       if (errno == EPIPE) {
		    sdPtr->state == SOCK_STATE_CLOSED;
		    return(SOCK_FD_CLOSED);
	       }
	       return(SOCK_FAILURE);
	  } else 
	       return(temp);
     }
}

Sock_Status Sock_ReadInt(Socket *sdPtr, int *i)
{
     Sock_Status status;
     
     status = Sock_RecvFull(sdPtr, (char *)i, sizeof(int));
     *i = ntohl(*i);
     return(status);
}

Sock_Status Sock_SendInt(Socket *sdPtr, int i)
{
     i = htonl(i);
     return(Sock_Send(sdPtr, (char *)&i, sizeof(int)));
}

Sock_Status Sock_ReadChar(Socket *sdPtr, char *c)
{
     return(Sock_RecvFull(sdPtr, c, sizeof(char)));
}

Sock_Status Sock_SendChar(Socket *sdPtr, char c)
{
     return(Sock_Send(sdPtr, (char *)&c, sizeof(char)));
}

Sock_Status Sock_ReadLong(Socket *sdPtr, long *l)
{
     Sock_Status status;
     
     status = Sock_RecvFull(sdPtr, (char *)l, sizeof(long));
     *l = ntohl(*l);
     return(status);
}

Sock_Status Sock_SendLong(Socket *sdPtr, long l)
{
     l = htonl(l);
     return(Sock_Send(sdPtr, (char *)&l, sizeof(long)));
}

Sock_Status Sock_ReadString(Socket *sdPtr, char *s)
{
     const int startingBufSize = 100;
     long bytesRead, currentRead;
     static char currentBufSize = -1;
     static char* buffer = NULL;

     if ((buffer = (char *)malloc(startingBufSize)) == NULL) {
          perror("SOCK_ReadString: Couldn't allocate initial buffer\n");
          return(SOCK_FAILURE);
     }
     currentBufSize = startingBufSize;
     bytesRead = 0;
     do {
          currentRead += Sock_RecvFull(sdPtr, buffer+bytesRead, sizeof(char));
          if (currentRead != sizeof(char)) {
               return(SOCK_FAILURE);
          }
          bytesRead += currentRead;
          if (bytesRead == currentBufSize) {
               /* grow the buffer */
               buffer = (char *)realloc(buffer, currentBufSize * 2);
               currentBufSize *= 2;
          }
     } while (buffer[bytesRead] != '\0');
     
     strcpy(s, buffer);
     free(buffer);
     return((Sock_Status)bytesRead);
}

Sock_Status Sock_SendString(Socket *sdPtr, char *s)
{
     return(Sock_Send(sdPtr, (char *)s, strlen(s)));
}

/*************************************************************************
  get/set routines
  *************************************************************************/

Sock_Status Sock_SetOption(Socket *sdPtr, OPTIONS option, int value)
{
     Sock_Status status=SOCK_SUCCESS;

     switch (option) {
#ifdef MULTICAST
     case TTL:
     {
	  int j = sizeof(value);
	  if (setsockopt(sdPtr->fd, IPPROTO_IP, IP_TTL,
			 (char *)&value,j) < 0) {
	       perror("Sock_Socket: ip_ttl");
	       status=SOCK_FAILURE;
	  } else 
	       sdPtr->ttl = value;
     }
	  break;
#endif
#ifdef TOS	 
     case IPTOS:
     {
	  int j = sizeof(value);
	  if (setsockopt(sdPtr->fd, IPPROTO_IP, IP_TOS,
			 (char *)&value,j) < 0) {
	       perror("Sock_Socket: ip_tos");
	       status=SOCK_FAILURE;
	  }
     }
	  break;
#endif
     case SEND_FLAG:
	  sdPtr->lFlags = value;
	  break;
     case MSS:
     {
	  int i, j = sizeof(value);
	  struct protoent *proto = getprotobyname("tcp");
	  if (sdPtr->protocol==SOCK_TCP) {
	       if (value < 500)
		    dprintf(109,"Sock_Setopt: low mss\n");
	       if (setsockopt(sdPtr->fd,proto->p_proto, 
			      TCP_MAXSEG,(char *)&value,j) < 0)
		    perror("Sock_SetOpt: set mss");
	       if (getsockopt(sdPtr->fd,proto->p_proto, 
			      TCP_MAXSEG,(char *)&i,&j) < 0)
		    perror("Sock_SetOpt: get mss");
	       dprintf(109,"Sock_Setopt: set mss to %d\n", i);
	  }
     }
	  break;
     case NODELAY: 
     {
	  int i,j = sizeof(i);
	  struct protoent *proto = getprotobyname("tcp");
	  i=1;
	  if (sdPtr->protocol==SOCK_TCP) {
	       if (value)
		    i = 1;
	       else 
		    i = 0;
	       if (setsockopt(sdPtr->fd,proto->p_proto, 
			      TCP_NODELAY,(char *)&i,j) < 0)
		    perror("Sock_Socket: tcp_nodelay");
	  }
     }
	  break;
     case PROTOCOL:
          dprintf(101, "Sock_SetOption: not implemented\n");
          status=SOCK_FAILURE;
          break;
     case TYPE:
          dprintf(101, "Sock_SetOption: not implemented\n");
          status=SOCK_FAILURE;
          break;
     case CLEAR_SOCKET_FLAG:
     {
          int val;
          val = fcntl(sdPtr->fd, F_GETFL, 0);
          if (val < 0)
               status = SOCK_FAILURE;
          else {
               val &= ~value;
               if (fcntl(sdPtr->fd, F_SETFL, val) < 0)
                    status = SOCK_FAILURE;
          }
     }
	  break;
     case SET_SOCKET_FLAG:
     {
          int val;
          val = fcntl(sdPtr->fd, F_GETFL, 0);
          if (val < 0)
               status = SOCK_FAILURE;
          else {
               val |= value;
               if (fcntl(sdPtr->fd, F_SETFL, val) < 0)
                    status = SOCK_FAILURE;
          }
     }
          break;
     case BUFFER_SIZE:
     {
          int k=0,l,m=0,n;

          l = sizeof(k);
          n = sizeof(m);
          setsockopt(sdPtr->fd, SOL_SOCKET,SO_SNDBUF,(char*)&value,
                     sizeof(value));
          getsockopt(sdPtr->fd,SOL_SOCKET,SO_SNDBUF,(char*)&k,&l);
          setsockopt(sdPtr->fd,SOL_SOCKET,SO_RCVBUF,(char*)&value,
                     sizeof(value));
          getsockopt(sdPtr->fd,SOL_SOCKET,SO_RCVBUF,(char*)&m,&n);
          dprintf(109, "Sock_setbuffersize: send buf %d, recv buf %d\n", k, m);
          sdPtr->lRecvBufSize = m;
          sdPtr->lXmitBufSize = k;
          if ((k == value) && (m == value))
               status = SOCK_SUCCESS;
          else
               status = SOCK_FAILURE;
     }
          break;
     case TIMEOUT:
          if (value == SOCK_NO_TIMEOUT) {
               sdPtr->toutPtr = NULL;
               dprintf(109, "no timeouts\n");
          }
          else {
               sdPtr->toutPtr = &(sdPtr->timeout);
               sdPtr->timeout.tv_sec = value/1000000;
               sdPtr->timeout.tv_usec = value -
                    (sdPtr->timeout.tv_sec * 1000000);
               dprintf(109, "timeout set to %d sec %d usec\n",
                       sdPtr->timeout.tv_sec, sdPtr->timeout.tv_usec );
          }
          break;
     case USE_CTRL_CHANNEL:
          sdPtr->useCtrl = value;
          break;
     case XMIT_RATE:
          sdPtr->xmitRate = value;
          break;
     case COMPATIBILITY:
          sdPtr->compatible = value;
          break;
     case BLOCKING:
          if (value==1) {
               if ((status = Sock_ClearFlag(sdPtr, O_NDELAY)) != SOCK_FAILURE)
		    sdPtr->block = SOCK_BLOCK;
          }
          else {
               if ((status = Sock_SetFlag(sdPtr, O_NDELAY)) != SOCK_FAILURE)
		    sdPtr->block = SOCK_NON_BLOCK;
          }
          break;
     case MOBILE:
          if (value==0)
               sdPtr->mobile = SOCK_NOT_MOBILE;
          else {
               sdPtr->mobile = SOCK_MOBILE;
               sdPtr->useCtrl = SOCK_USE_CTRL;
          }
          break;
     case BROADCAST:
          if (value==0) {
               char *ipAddr, *newIPAddr;
               Port port;
               int i,k=-1;

               if (sdPtr->protocol == SOCK_TCP)
                    return(SOCK_FAILURE);
               if (sdPtr->broadcast == SOCK_NOT_BROADCAST)
                    return(SOCK_SUCCESS);
               if (!((sdPtr->state==SOCK_STATE_CONNECTED) ||
                     (sdPtr->state==SOCK_STATE_ACCEPTED)))
                    return(SOCK_SUCCESS);
               if (sdPtr->rHost == NULL)
                    return(SOCK_FAILURE);
               port = Sock_GetRemotePort(sdPtr);
               status=Sock_SetRemoteByName(sdPtr, sdPtr->rHost->remoteName, 
					   port);
               sdPtr->broadcast = SOCK_NOT_BROADCAST;
          }
          else {
               Address *rSinPtr;
               char *ipAddr, *newIPAddr;
               Port port;

               if (sdPtr->protocol == SOCK_TCP)
                    return(SOCK_FAILURE);
               if (sdPtr->broadcast == SOCK_BROADCAST)
                    return(SOCK_SUCCESS);
               sdPtr->broadcast = SOCK_BROADCAST;
               if (!((sdPtr->state==SOCK_STATE_CONNECTED) ||
                     (sdPtr->state==SOCK_STATE_ACCEPTED)))
                    return(SOCK_SUCCESS);
               /* XXXX fix to get correct addr */
               rSinPtr = Sock_GetRemoteAddr(sdPtr);
               port = Sock_GetRemotePort(sdPtr);
               ipAddr = inet_ntoa(rSinPtr->sin_addr);
               newIPAddr = Sock_BroadcastAddr(ipAddr);
               status = Sock_SetRemoteByName(sdPtr, newIPAddr, port);
          }
          break;
     default:
          dprintf(101, "Sock_SetOption: invalid option %d\n", option);
          status = SOCK_FAILURE;
     }
     return(status);
}

int Sock_GetOption(Socket *sdPtr, OPTIONS option)
{
     int value;

     switch (option) {
#ifdef MULTICAST
     case TTL:
     {
	  int j = sizeof(value);
	  if (getsockopt(sdPtr->fd, IPPROTO_IP, IP_TTL,
			 (char *)&value,&j) < 0) {
	       perror("Sock_Socket: ip_ttl");
	       value=SOCK_FAILURE;
	  }
     }
	  break;
#endif
#ifdef TOS	 
     case IPTOS:
     {
	  int j = sizeof(value);
	  if (getsockopt(sdPtr->fd, IPPROTO_IP, IP_TOS,
			 (char *)&value,&j) < 0) {
	       perror("Sock_Socket: ip_tos");
	       value=SOCK_FAILURE;
	  }
     }
	  break;
#endif
     case SEND_FLAG:
	  value = sdPtr->lFlags;
	  break;
     case NODELAY:
     {
	  int j = sizeof(value);
	  if (getsockopt(sdPtr->fd,IPPROTO_TCP,
			 TCP_NODELAY,(char *)&value,&j) < 0)
		    perror("Sock_Socket: tcp_nodelay");
     }
	  break;
     case PROTOCOL:
          value=sdPtr->protocol;
          break;
     case TYPE:
          value=sdPtr->type;
          break;
     case GET_SOCKET_FLAG:
          value = fcntl(sdPtr->fd, F_GETFL, 0);
          break;
     case BUFFER_SIZE:
          value=sdPtr->lRecvBufSize;
          break;
     case TIMEOUT:
          if (sdPtr->toutPtr == NULL) {
               value = SOCK_NO_TIMEOUT;
          }
          else {
               value=sdPtr->timeout.tv_sec * 1000000 + sdPtr->timeout.tv_usec;
          }
          break;
     case USE_CTRL_CHANNEL:
          value = sdPtr->useCtrl;
          break;
     case XMIT_RATE:
          value = sdPtr->xmitRate;
          break;
     case COMPATIBILITY:
          value = sdPtr->compatible;
          break;
     case BLOCKING:
          value = sdPtr->block;
          break;
     case MOBILE:
          value = sdPtr->mobile;
          break;
     case BROADCAST:
          value = sdPtr->broadcast;
          break;
     default:
          dprintf(101, "Sock_GetOption: invalid option %d\n", option);
          value = SOCK_FAILURE;
     }
     return(value);
}

int Sock_GetFD(Socket *sdPtr)
{
     return(sdPtr->fd);
}

char *Sock_GetRemoteName(Socket *sdPtr)
{
     char *retStr;

     retStr = (char *)malloc(strlen(sdPtr->rHost->remoteName));
     strcpy(retStr,sdPtr->rHost->remoteName);
     return(retStr);
}

void Sock_SetBer(Socket *sdPtr, float val)
{
     sdPtr->berState.ErrorGen = Sock_PoissonError;
     sdPtr->berState.errorProb[0] = val;
     sdPtr->berState.bytesToNextError = 0;
     sdPtr->berState.bitsDropped = 0;
     sdPtr->berState.totalBits = 0;
}

float Sock_GetBer(Socket *sdPtr) 
{
     return(sdPtr->berState.errorProb[0]);
}

Address *Sock_GetRemoteAddr(Socket *sdPtr)
{
     Address *retAddr;

     retAddr = (Address *)malloc(sizeof(Address));
     memcpy(retAddr, &(sdPtr->rSin), sizeof(Address));
     return(retAddr);
}

Address *Sock_GetLocalAddr(Socket *sdPtr)
{
     Address *retAddr;

     retAddr = (Address *)malloc(sizeof(Address));
     memcpy(retAddr, &(sdPtr->lSin), sizeof(Address));
     return(retAddr);
}

Address *Sock_GetLastSenderAddr(Socket *sdPtr)
{
     Address *retAddr;

     retAddr = (Address *)malloc(sizeof(Address));
     memcpy(retAddr, &(sdPtr->lastRSin), sizeof(Address));
     return(retAddr);
}

Port Sock_GetLocalPort(Socket *sdPtr)
{
     return(sdPtr->localPort);
}

Port Sock_GetRemotePort(Socket *sdPtr)
{
     return(sdPtr->remotePort);
}

Port Sock_GetTCPCtrlPort(Socket *sdPtr)
{
     if (sdPtr->tcpCtrlChannel != NULL)
	  return(sdPtr->tcpCtrlChannel->localPort);
     else
	  return(SOCK_FAILURE);
}

Port Sock_GetUDPCtrlPort(Socket *sdPtr)
{
     if (sdPtr->udpCtrlChannel != NULL)
	  return(sdPtr->udpCtrlChannel->localPort);
     else
	  return(SOCK_FAILURE);
}

Sock_Status Sock_SetRemoteByName(Socket *sdPtr, char *hostname, Port port)
{
     char *t2;

     /* get remote host address */
     dprintf(109,"setting remote to %s %d\n", hostname, port);

     Sock_NameToSin(hostname, port, &(sdPtr->rSin));
     /* set up socket descriptor */
     t2 = sdPtr->rHost->remoteName;
     sdPtr->rHost->remoteName = (char *) malloc(strlen(hostname));
     strcpy(sdPtr->rHost->remoteName, hostname);
     if (t2 != NULL)
	  free(t2);
     sdPtr->remotePort = port;
     return(SOCK_SUCCESS);
}

Sock_Status Sock_SetRemoteByAddr(Socket *sdPtr, Address *rSinPtr)
{
     char *t2;

     if (rSinPtr != &(sdPtr->rSin))
	  memcpy(&(sdPtr->rSin), rSinPtr, sizeof(sdPtr->rSin));
     
     t2 = sdPtr->rHost->remoteName;
     Sock_SinToName(&(sdPtr->rHost->remoteName), &(sdPtr->remotePort),rSinPtr);
     if (t2 != NULL)
	  free(t2);
     return(SOCK_SUCCESS);
}

Sock_Status Sock_AddDestination(Socket *sdPtr, char *hostname)
{
     Socket *newSdPtr;
     Sock_Status status;
     int i=0;
     /*
	
	if (sdPtr->protocol == SOCK_TCP)
	return(SOCK_FAILURE);
	if (sdPtr->broadcast == SOCK_NOT_BROADCAST)
	return(SOCK_FAILURE);
	if (!((sdPtr->state== SOCK_STATE_CONNECTED) ||
	(sdPtr->state== SOCK_STATE_ACCEPTED)))
	return(SOCK_FAILURE);
	
	newSdPtr = Sock_Socket(sdPtr->protocol);
	status = Sock_Connect(newSdPtr, hostname, sdPtr->remotePort);
	
	if (status == SOCK_FAILURE) {
	dprintf(109, "couldn't add destination %s\n", hostname);
	return(SOCK_FAILURE);
	}
	
	for (i=0; i < SOCK_MAX_MULTICAST_HOSTS; i++)
	if (sdPtr->activeHosts[i] == 0)
	break;
	
	strcpy(sdPtr->remoteName[i], newSdPtr->rHost->remoteName);
	sdPtr->ctrlChannel[i] = newSdPtr->ctrlChannel[0];
	sdPtr->activeHosts[i] = 1;
	*/
     dprintf(109, "added destination %s as %d\n", hostname, i);
     return(status);
}

Sock_Status Sock_MergeDestinations(Socket *sdPtr1, Socket *sdPtr2)
{
     int i;
     /*
	if (sdPtr1->protocol == SOCK_TCP)
	return(SOCK_FAILURE);
	if (!((sdPtr1->state==SOCK_STATE_CONNECTED) ||
	(sdPtr1->state==SOCK_STATE_ACCEPTED)))
	return(SOCK_FAILURE);
	
	for (i=0; i < SOCK_MAX_MULTICAST_HOSTS; i++)
	if (sdPtr1->activeHosts[i] == 0)
	break;
	
	strcpy(sdPtr1->remoteName[i], sdPtr2->rHost->remoteName);
	sdPtr1->ctrlChannel[i] = sdPtr2->ctrlChannel[0];
	sdPtr1->activeHosts[i] = 1;
	*/
     dprintf(109, "merged destination %s as %d\n",sdPtr2->rHost->remoteName , i);
     return(SOCK_SUCCESS);
}

Sock_Status Sock_DelDestination(Socket *sdPtr, char *hostname)
{
     int i,j=0;

     /*     for (i=0; i < SOCK_MAX_MULTICAST_HOSTS; i++)
	    if (sdPtr->activeHosts[i] == 1)
	    if (strcmp(sdPtr->remoteName[i], hostname))
	    break;
	    
	    dprintf(109, "deleting host %d = %s\n", i, hostname);
	    
	    Sock_CtrlClose(sdPtr->ctrlChannel[i], sdPtr->localName);
	    Sock_Close(sdPtr->ctrlChannel[i]);
	    sdPtr->activeHosts[i] = 0;
	    */
}

