/*
 * Copyright (c) 1995, 1996, 1997
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* 
 * Functions for advice handling and state reporting.
 * Original version written by Srinivasan Seshan (srini@watson.ibm.com)
 * modified heavily and released by Mark Stemm (stemm@cs.berkeley.edu)
 *
 * Daedalus group, Dept. of EECS, U.C. Berkeley.
 * Contact: stemm@cs.berkeley.edu, net@daedalus.cs.berkeley.edu
 */

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <signal.h>
#include <unistd.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/in_var.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ip_mobile.h"
#include "easysocklib.h"
#include "option.h"
#include "dbgprintlib.h"

#include "beacon.h"
#include "decap.h"
#include "sample.h"
#include "newroute.h"
#include "vroute.h"

#include <advice_state.h>
#include <handoffctlr.h>

extern Socket *adviceRepl_sdPtr;
extern Socket *stateRepl_sdPtr;

/* Convenience macro for ProcessAdviceRequest and ProcessStateRequest */
#define SEND_ADV_REPLY	{ Advice_MsgOpcode(&replyMsg) = Advice_MsgOpcode(aMsg) | REPLY_MASK; if (Sock_Send(adviceRepl_sdPtr, (char *) &replyMsg, sizeof(replyMsg)) == SOCK_FAILURE) { perror("Sock_Send()"); return; } }

/*
 * Change the state used to make decisions about handoffs to networks and
 * basestation.
 *   curiPtr: pointer to the current interface
 *   newIfacename: name of the new interface to handoff to
 *   newBSaddr: pointer to address of the new BS to handoff to.
 * This may also do a handoff if the advice message requires it.
 */
void
ChangeHandoffState(int type, RouteState * sysStatePtr, IFace * curiPtr, char *newIfacename, struct in_addr * newBSaddr)
{
	
	IFace  *newiPtr;
	BSState *newBSPtr;
	int     i, j;
	
	switch (type) {
	case USE_THIS_NETWORK:
		/* find the pointer to the new interface */
		for (i = 0; i < sysStatePtr->numIfaces; i++) {
			if (!strcmp(sysStatePtr->iface[i]->ifname, newIfacename))
				break;
		}
		if (i >= sysStatePtr->numIfaces) {
			dprintf(5, "ChangeHandoffState(): Unknown interface %s: handoff request failed\n", newIfacename);
			return;
		}
		newiPtr = sysStatePtr->iface[i];
		/* find a basestation to handoff to on this network */
		for (j = 0; j < newiPtr->numBS; j++)
			if ((newiPtr->BSPtr[j]->state != BS_INACTIVE) && (!newiPtr->BSPtr[j]->dontUse))
				break;
		if (j >= newiPtr->numBS) {
			dprintf(5, "Error: USE_THIS_NETWORK: unable to find any basestation to handoff to; reseting all dontUse flags\n");
			CancelUserCmds(sysStatePtr);
			return;
		}
		newBSPtr = newiPtr->BSPtr[j];
		/* set the dontUse flag for all interfaces except newiPtr */
		for (i = 0; i < sysStatePtr->numIfaces; i++)
			sysStatePtr->iface[i]->dontUse = 1;
		newiPtr->dontUse = 0;
		break;
		
	case DONT_USE_THIS_NETWORK:
		/* find the pointer to the specified interface */
		for (i = 0; i < sysStatePtr->numIfaces; i++) {
			if (!strcmp(sysStatePtr->iface[i]->ifname, newIfacename))
				break;
		}
		if (i >= sysStatePtr->numIfaces) {
			dprintf(5, "ChangeHandoffState(): DONT_USE_THIS NETWORK: Unknown interface %s: request failed\n", newIfacename);
			return;
		}
		sysStatePtr->iface[i]->dontUse = 1;
		/* if the specified interface is the same as curiPtr, find a
		 * new interface and basestation to handoff to */
		if (curiPtr == sysStatePtr->iface[i]) {
			for (i = 0; i < sysStatePtr->numIfaces; i++) {
				if (sysStatePtr->iface[i]) {	/* found an interface;
								 * look for an active BS */
					newiPtr = sysStatePtr->iface[i];
					if (newiPtr->dontUse)
						continue;
					for (j = 0; j < newiPtr->numBS; j++) {
						if (newiPtr->BSPtr[j]) {
							if ((newiPtr->BSPtr[j]->state != BS_INACTIVE) && (!newiPtr->BSPtr[j]->dontUse))
								break;
						}
					}
					if (j < newiPtr->numBS) {	/* found a BS! */
						newBSPtr = newiPtr->BSPtr[j];
						break;
					}
				}
			}
			if (i >= sysStatePtr->numIfaces) {
				dprintf(5, "ChangeHandoffState(): DONT_USE_THIS_NETWORK: Error: unable to find any basestation to handoff to; resetting all dontUse flags\n");
				CancelUserCmds(sysStatePtr);
				return;
			}
			break;
		} else		/* no handoff to be done */
			return;
		
		
	case USE_THIS_BASESTATION:
		/* find the pointer to the new BS and the new interface */
		for (i = 0; i < sysStatePtr->numIfaces; i++) {
			newiPtr = sysStatePtr->iface[i];
			for (j = 0; j < newiPtr->numBS; j++) {
				if (newiPtr->BSPtr[j]->wlAddr.s_addr == newBSaddr->s_addr)
					break;
			}
			if (j < newiPtr->numBS) {
				newBSPtr = newiPtr->BSPtr[j];
				break;
			}
		}
		if (i >= sysStatePtr->numIfaces) {
			dprintf(5, "ChangeHandoffState(): USE_THIS_BASESTATION: Unknown basestation %x: handoff request failed; resetting all dontUse flags\n", newBSaddr->s_addr);
			CancelUserCmds(sysStatePtr);
			return;
		}
		/* set the dontUse flag for all basestations except newBSPtr */
		for (i = 0; i < sysStatePtr->numIfaces; i++)
			for (j = 0; j < sysStatePtr->iface[i]->numBS; j++)
				sysStatePtr->iface[i]->BSPtr[j]->dontUse = 1;
		newBSPtr->dontUse = 0;
		break;
		
	case DONT_USE_THIS_BASESTATION:
		/* find the pointer to the specified BS */
		for (i = 0; i < sysStatePtr->numIfaces; i++) {
			for (j = 0; j < sysStatePtr->iface[i]->numBS; j++) {
				if (sysStatePtr->iface[i]->BSPtr[j]->wlAddr.s_addr == newBSaddr->s_addr)
					break;
			}
			if (j < sysStatePtr->iface[i]->numBS)
				break;
		}
		if (i >= sysStatePtr->numIfaces) {
			dprintf(5, "ChangeHandoffState(): DONT_USE_THIS_BASESTATION: Unknown basestation %x: request failed\n", newBSaddr->s_addr);
			return;
		}
		sysStatePtr->iface[i]->BSPtr[j]->dontUse = 1;
		/* if the specified basestation is the same as
		 * curiPtr->curBSPtr, find a new basestation to handoff to */
		if (curiPtr->curBSPtr == sysStatePtr->iface[i]->BSPtr[j]) {
			for (i = 0; i < sysStatePtr->numIfaces; i++) {
				if (sysStatePtr->iface[i]) {	/* found an interface;
								 * look for active BS */
					newiPtr = sysStatePtr->iface[i];
					if (newiPtr->dontUse)
						continue;
					for (j = 0; j < newiPtr->numBS; j++) {
						if (newiPtr->BSPtr[j]) {
							if ((newiPtr->BSPtr[j]->state != BS_INACTIVE) && (!newiPtr->BSPtr[j]->dontUse))
								break;
						}
					}
					if (j < newiPtr->numBS) {	/* found a BS! */
						newBSPtr = newiPtr->BSPtr[j];
						break;
					}
				}
			}
			if (i >= sysStatePtr->numIfaces) {
				dprintf(5, "ChangeHandoffState(): Error: unable to find any basestation to handoff to; resetting all dontUse flags\n");
				CancelUserCmds(sysStatePtr);
				return;
			}
			break;
		} else		/* no handoff to be done */
			return;
		
		
	default:
		dprintf(5, "ChangeHandoffState(): request of unknown type: %d\n", type);
	}
	
	/* now actually do the handoff if the new BS is indeed different from
	 * the current one */
	if (curiPtr->curBSPtr != newBSPtr)
		DoHandoff(sysStatePtr, curiPtr, newiPtr, newBSPtr, ADVICE);
	
}


/*
 * Process an advice message from the user controller.
 *   advicemsg: the message received from the user controller
 *   sysSatePtr: pointer to the network state date structures
 *   curiPtr: pointer to the current interface
 */
void
ProcessAdviceRequest(char *advicemsg, RouteState * sysStatePtr, IFace * curiPtr)
{
	AdviceMsg *aMsg = (AdviceMsg *) advicemsg;
	AdviceMsg replyMsg;
	
	dprintf(25, "In ProcessAdviceRequest(): opcode=%x\n", Advice_MsgOpcode(aMsg));
	
	
	switch (Advice_MsgOpcode(aMsg)) {
		
	case USE_THIS_NETWORK:{/* use a particular interface */
		
		dprintf(25, "In ProcessAdviceRequest(): ifname=%s\n", Advice_MsgIf(aMsg));
		ChangeHandoffState(USE_THIS_NETWORK, sysStatePtr, curiPtr, Advice_MsgIf(aMsg), NULL);
		/* send back a reply */
		SEND_ADV_REPLY;
		break;
	}
	
	case USE_ANY_NETWORK:{	/* use any interface */
		int     i;
		
		/* reset dontUse flag for all interfaces */
		for (i = 0; i < sysStatePtr->numIfaces; i++)
			sysStatePtr->iface[i]->dontUse = 0;
		/* send back a reply */
		SEND_ADV_REPLY;
		break;
	}
	
	case DONT_USE_THIS_NETWORK:{
		/* don't use this particular interface */
		
		dprintf(25, "In ProcessAdviceRequest(): ifname=%s\n", Advice_MsgIf(aMsg));
		/* call ChangeHandoffState(); an actual handoff is done
		 * only if we are currently using the interface that
		 * is disallowed */
		ChangeHandoffState(DONT_USE_THIS_NETWORK, sysStatePtr, curiPtr, Advice_MsgIf(aMsg), NULL);
		/* send back a reply */
		SEND_ADV_REPLY;
		break;
	}
	
	case USE_THIS_BASESTATION:{
		/* use a particular basestation */
		
		dprintf(25, "In ProcessAdviceRequest(): bsaddr=%x\n", (Advice_MsgBs(aMsg)).s_addr);
		ChangeHandoffState(USE_THIS_BASESTATION, sysStatePtr, curiPtr, NULL, &(Advice_MsgBs(aMsg)));
		/* send back a reply */
		SEND_ADV_REPLY;
		break;
	}
	
	case USE_ANY_BASESTATION:{
		int     i, j;
		
		/* reset dontUse flags for all basestations */
		for (i = 0; i < sysStatePtr->numIfaces; i++)
			for (j = 0; j < sysStatePtr->iface[i]->numBS; j++)
				sysStatePtr->iface[i]->BSPtr[j]->dontUse = 0;
		/* send back a reply */
		SEND_ADV_REPLY;
		break;
	}
	
	case DONT_USE_THIS_BASESTATION:{
		
		dprintf(25, "In ProcessAdviceRequest(): bsaddr=%x\n", (Advice_MsgBs(aMsg)).s_addr);
		/* call ChangeHandoffState(); an actual handoff is done
		 * only if we are currently using a basestation that
		 * is disallowed */
		ChangeHandoffState(DONT_USE_THIS_BASESTATION, sysStatePtr, curiPtr, NULL, &(Advice_MsgBs(aMsg)));
		/* send back a reply */
		SEND_ADV_REPLY;
		break;
	}
	default:
		dprintf(25, "Unrecognized opcode %d for advice msg--ignoring\n", Advice_MsgBs(aMsg));
		break;				
	}
	
	return;
}

#define SEND_STATE_MSG(sock) { int size, i;  size=Handoffctlr_MsgToString(replymsgstring, &replyMsg);	dprintf(25, "Sending...\n"); for(i=0; i<size; i++){ dprintf(25, "%x ", replymsgstring[i]);} if (Sock_Send(sock, replymsgstring, size) == SOCK_FAILURE) { perror("Sock_Send()"); return; }  }

#define SEND_STATE_REPLY { Handoffctlr_MsgOpcode(&replyMsg) = Handoffctlr_MsgOpcode(&stateMsg) | REPLY_MASK; SEND_STATE_MSG(stateRepl_sdPtr); }

#define COMPLAIN_IF_NUMBODY_NOT(i) { if ((Handoffctlr_MsgNumBody(&stateMsg)) != i) { dprintf(5, "ProcessStateRequest(): num_body=%d (!=%d); aborting!\n", Handoffctlr_MsgNumBody(&stateMsg), i); return; }}


/*		
 * Process a message from the user controller.
 *   advicemsg: the message received from the user controller
 *   sysSatePtr: pointer to the network state date structures
 *   curiPtr: pointer to the current interface
 */
void
ProcessStateRequest(char *statemsg, RouteState * sysStatePtr, IFace * curiPtr)
{
	HandoffctlrMsg stateMsg, replyMsg;
	char    replymsgstring[1000];
	int     i, j;
	IFace  *iPtr;
	
	Handoffctlr_StringToMsg(&stateMsg, statemsg);
	dprintf(25, "In ProcessStateRequest(): opcode=%x\n", Handoffctlr_MsgOpcode(&stateMsg));
	
	switch (Handoffctlr_MsgOpcode(&stateMsg)) {
		
	case LIST_ALL_NETWORKS:
		/* list all networks */
		
		COMPLAIN_IF_NUMBODY_NOT(0);
		
		Handoffctlr_MsgNumBody(&replyMsg)=sysStatePtr->numIfaces;
		for (i = 0; i < sysStatePtr->numIfaces; i++) {
			strcpy(Handoffctlr_MsgIf(&replyMsg, i), sysStatePtr->iface[i]->ifname);
		}
		
		dprintf(25, "List all networks: Num=%d  ", Handoffctlr_MsgNumBody(&replyMsg));
		for (i = 0; i < Handoffctlr_MsgNumBody(&replyMsg); i++)
			dprintf(25, "%s ", Handoffctlr_MsgIf(&replyMsg, i));
		dprintf(25, "\n");
		
		/* send back a reply */
/*		SEND_STATE_REPLY;*/
		break;
		
	case LIST_ALL_BASESTATIONS:
		/* list all basestations on the specified
		 * interface. */
		
		COMPLAIN_IF_NUMBODY_NOT(1);
		
		for (i = 0; i < sysStatePtr->numIfaces; i++) {
			iPtr = sysStatePtr->iface[i];
			if (!strcmp(Handoffctlr_MsgIf(&stateMsg, 0), iPtr->ifname))
				break;
		}
		
		if (i >= sysStatePtr->numIfaces) {
			dprintf(5, "ProcessAdviceRequest(): LIST_ALL_BASESTATIONS: unknown interface %s\n", Handoffctlr_MsgIf(&stateMsg, 0));
			return;
		}
		
		Handoffctlr_MsgNumBody(&replyMsg)=iPtr->numBS;
		
		for (j = 0; j < iPtr->numBS; j++) {
			(Handoffctlr_MsgAddr(&replyMsg, j)).s_addr = iPtr->BSPtr[j]->wlAddr.s_addr;
		}
		
		dprintf(25, "List all basestations: Num=%d  ", Handoffctlr_MsgNumBody(&replyMsg));
		for (j = 0; j < Handoffctlr_MsgNumBody(&replyMsg); j++)
			dprintf(25, "%x ", Handoffctlr_MsgAddr(&replyMsg, j).s_addr);
		dprintf(25, "\n");
		
		/* send back a reply */
/*		SEND_STATE_REPLY;*/
		break;
		
	case LIST_ACTIVE_NETWORK:
		/* return the active network */
		
		COMPLAIN_IF_NUMBODY_NOT(0);
		
		Handoffctlr_MsgNumBody(&replyMsg)=1;
		
		strcpy(Handoffctlr_MsgIf(&replyMsg, 0), sysStatePtr->curiPtr->ifname);
		
		dprintf(25, "List active network: Num=%d Iface: %s\n", Handoffctlr_MsgNumBody(&replyMsg), Handoffctlr_MsgIf(&replyMsg, 0));
		
		/* send back a reply */
/*		SEND_STATE_REPLY;*/
		break;
		
	case LIST_ACTIVE_BASESTATION:
		/* return the active basestation */
		
		COMPLAIN_IF_NUMBODY_NOT(0);
		
		iPtr = sysStatePtr->curiPtr;
		Handoffctlr_MsgAddr(&replyMsg, 0).s_addr = iPtr->curBSPtr->wlAddr.s_addr;
		
		dprintf(25, "List active basestations: Num=%d Addr: %x\n", Handoffctlr_MsgNumBody(&replyMsg), Handoffctlr_MsgAddr(&replyMsg, 0).s_addr);
		
		/* send back a reply */
/*		SEND_STATE_REPLY;*/
		break;
		
	default:
		dprintf(25, "Unrecognized opcode %d, ignoring state message\n", Handoffctlr_MsgOpcode(&stateMsg));
		break;
	}
	
	return;
}

/*
 * Reset the dontUse flags corresponding to all interfaces and basestations.
 */
void
CancelUserCmds(RouteState * sysStatePtr)
{
	int     i, j;
	IFace  *iPtr;
	
	for (i = 0; i < sysStatePtr->numIfaces; i++) {
		iPtr = sysStatePtr->iface[i];
		iPtr->dontUse = 0;
		for (j = 0; j < iPtr->numBS; j++)
			iPtr->BSPtr[j]->dontUse = 0;
	}
}


