/*
 * Copyright (c) 1995, 1996
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* 
* User-level decapsulation daemon for Daedalus Project's handoff software.
 * Original version written by Srinivasan Seshan (srini@watson.ibm.com)
 * modified heavily and released by Mark Stemm (stemm@cs.berkeley.edu)
 *
 * Daedalus group, Dept. of EECS, U.C. Berkeley.
 * Contact: stemm@cs.berkeley.edu, net@daedalus.cs.berkeley.edu
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <net/if.h>
#include <net/if_types.h>
#include <netinet/in.h>
#include <netinet/in_var.h>

#include <newroute.h>
#include <conf.h>

#include "ip_mobile.h"
#include "easysocklib.h"
#include "option.h"
#include "dbgprintlib.h"

#include "decap.h"

int debug = 0;
Port decapPort = BSR_PORT;
char *wiName = "ajanta";
char *wlNames = "ajanta-wl";
char *confFileName = "/etc/daedalus.conf";

Option optionArray[] = {
	{OPT_INT, "debug", (char*)&debug,
	 "Turn on debugging (higher value -> more debugging)."},
	{OPT_STRING, "confFileName", (char *)&confFileName,
	 "Configuration file for paramaters (default /etc/daedalus.conf)."},
	{OPT_INT, "decapPort", (char *)&decapPort,
	 "decapPort"},
	{OPT_STRING, "wiName", (char *)&wiName,
	 "host if addr on fixed net"},
	{OPT_STRING, "wlNames", (char *)&wlNames,
	 "wireless names for bs"}
};

void NewFDSet(DecapdState *eStatePtr)
{
	int i;
	SocketList *tempPtr;
	
	FD_ZERO(&eStatePtr->fdvar);
	FD_SET(Sock_GetFD(eStatePtr->sdPtr), &eStatePtr->fdvar);
	for (tempPtr=eStatePtr->connList.nextSock; tempPtr!=NULL; tempPtr=tempPtr->nextSock){
		FD_SET(Sock_GetFD(tempPtr->sock), &eStatePtr->fdvar);
	}
}

void ClearKernelState()
{
	struct ip_mobreq req;
	int sockfd;
	int i, j;
	
	sockfd= socket(AF_INET, SOCK_STREAM, 0);
	i = 1;
	j = sizeof(i);
	if (setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY,(char *)&i,j) < 0)
		perror("Creating socket for kernel-level state reset:");
	
	if (setsockopt(sockfd, IPPROTO_IP, IP_CLR_ALL, (char *)&req, sizeof(req))<0){
		perror("Trying to reset kernel-level state:");
	}
}

void ClearKernelStateExit()
{
	ClearKernelState();
	exit(0);
}

void DisableMobile(DecapdState *eStatePtr, BSReq *bsreq)
{
	struct ip_mobreq req;
	int j = sizeof(req);
	int option;

	char ln[100];
	char rn[100];
	char mn[100];
	
	strcpy(ln, inet_ntoa(bsreq->locAddr));
	strcpy(rn, inet_ntoa(bsreq->realAddr));
	strcpy(mn, inet_ntoa(bsreq->multiAddr));
	
	dprintf(5, "message is: disable mobile real name=%s, local name=%s, multicast name=%s\n", rn,ln,mn);

	/* Delete the host route to the mobile */
	dprintf(10,"del route\n");
	DelHostRoute(rn, ln, eStatePtr->routes);
		
	/* Modify the kernel-level state: stop forwarding */
	option = IP_DIS_UNENCAP;
	req.hdrOnly = bsreq->hdrOnly;
	req.imr_multi_addr.inm_addr.s_addr = bsreq->multiAddr.s_addr;
	req.imr_multicast_ttl = DEF_TTL;
	req.imr_ifaddr.s_addr = eStatePtr->wiAddr.s_addr;
	req.imr_ins_addr.s_addr = bsreq->realAddr.s_addr;
	if (setsockopt(eStatePtr->ctrlFD,IPPROTO_IP, option,(char *)&req,j) < 0){
		perror("setsockopt in DisableMobile()");
	}
}

void EnableMobile(DecapdState *eStatePtr, BSReq *bsreq)
{
	struct ip_mobreq req;
	int j = sizeof(req);
	int option;
	char ln[100];
	char rn[100];
	char mn[100];
	
	strcpy(ln, inet_ntoa(bsreq->locAddr));
	strcpy(rn, inet_ntoa(bsreq->realAddr));
	strcpy(mn, inet_ntoa(bsreq->multiAddr));
	
	dprintf(5, "message is: fwd mobile real name=%s, local name=%s, multicast name=%s\n", rn,ln,mn);

	/* Add a host route to the real name of the mobile via the local name of the mobile */
	dprintf(10,"add route\n");
	DelHostRoute(rn, ln, eStatePtr->routes);
	AddHostRoute(rn, ln, eStatePtr->routes);
	
	/* Change the kernel-level state: start forwarding to this mobile */
	option = IP_ENA_UNENCAP;
	req.imr_multi_addr.inm_addr.s_addr = bsreq->multiAddr.s_addr;
	req.imr_multicast_ttl = DEF_TTL;
	req.imr_ifaddr.s_addr = eStatePtr->wiAddr.s_addr;
	req.imr_ins_addr.s_addr = bsreq->realAddr.s_addr;
	memcpy(&req.ids, &bsreq->ids, sizeof(bsreq->ids));
	
	if (setsockopt(eStatePtr->ctrlFD,IPPROTO_IP, option,(char *)&req,j) < 0){
		perror("setsockopt in EnableMobile()");
	}
	
}

void AddMobile(DecapdState *eStatePtr, BSReq *bsreq)
{
	struct ip_mobreq req;
	int j = sizeof(req);
	int option;
	char ln[100];
	char rn[100];
	char mn[100];

	strcpy(ln, inet_ntoa(bsreq->locAddr));
	strcpy(rn, inet_ntoa(bsreq->realAddr));
	strcpy(mn, inet_ntoa(bsreq->multiAddr));
	
	dprintf(5, "message is: add mobile real name=%s, local name=%s, multicast name=%s\n",
		rn, ln, mn);

	/* Add a host route to the real name of the mobile via the local name of the mobile */
	dprintf(10,"add route\n");
	DelHostRoute(rn, ln, eStatePtr->routes);
	AddHostRoute(rn, ln, eStatePtr->routes);

	/* Change the kernel-level state: add an entry for this mobile */
	option = IP_ADD_UNENCAP;
	req.imr_multi_addr.inm_addr.s_addr = bsreq->multiAddr.s_addr;
	req.imr_multicast_ttl = DEF_TTL;
	req.imr_ifaddr.s_addr = eStatePtr->wiAddr.s_addr;
	req.imr_ins_addr.s_addr = bsreq->realAddr.s_addr;
	if (setsockopt(eStatePtr->ctrlFD,IPPROTO_IP, option,(char *)&req,j) < 0){
		perror("setsockopt in AddMobile()");
	}
}		


void DelMobile(DecapdState *eStatePtr, BSReq *bsreq) 
{
	struct ip_mobreq req;
	int j = sizeof(req);
	int option;
	char ln[100];
	char rn[100];
	char mn[100];
	
	strcpy(ln, inet_ntoa(bsreq->locAddr));
	strcpy(rn, inet_ntoa(bsreq->realAddr));
	strcpy(mn, inet_ntoa(bsreq->multiAddr));
	
	dprintf(5, "message is: del mobile real name=%s, local name=%s, multicast name=%s\n", rn,ln,mn);
	
	/* Delete the host route to the mobile */
	dprintf(10,"del route\n");
	DelHostRoute(rn, ln, eStatePtr->routes);

	/* Delete the entry in the kernel-level state */
	dprintf(10,"del unencap\n");
	option = IP_DEL_UNENCAP;
	req.imr_multi_addr.inm_addr.s_addr = inet_addr(mn);
	req.imr_multicast_ttl = DEF_TTL;
	req.imr_ifaddr.s_addr = eStatePtr->wiAddr.s_addr;
	req.imr_ins_addr.s_addr = inet_addr(rn);
	if (setsockopt(eStatePtr->ctrlFD,IPPROTO_IP,option,(char *)&req,j) < 0)
		perror("setsockopt in DelMobile():");

}

void ForwardDecapMsg(BSReq *bsreq)
{
	Socket *temp;

	dprintf(20, "Forwarding to %s\n", inet_ntoa(bsreq->recipientAddr));
	
	temp = Sock_Socket(SOCK_TCP);
	Sock_SetCompatible(temp, 1);
	Sock_SetBufferSize(temp, 16384);
	if (Sock_Connect(temp, inet_ntoa(bsreq->recipientAddr), decapPort) == SOCK_FAILURE) {
		dprintf(5, "error connecting to bs %s\n", inet_ntoa(bsreq->recipientAddr));
		return;
	}
	if (Sock_Send(temp, (char *) bsreq, sizeof(BSReq)) == SOCK_FAILURE) {
		dprintf(5, "error sending to bs %s\n", inet_ntoa(bsreq->recipientAddr));
	}

	Sock_Close(temp);
	Sock_Free(temp);
}

int ProcessDecapMsg(Socket *rdPtr, DecapdState *eStatePtr)
{
	BSReq bsreq;

	/* Read the request */
	if (Sock_RecvFull(rdPtr,(char *)&bsreq,sizeof(bsreq))==SOCK_FD_CLOSED){
		return(-1);
	}

	/* If the recipient isn't us, then forward the request to the recipient */
	if (bsreq.recipientAddr.s_addr!=eStatePtr->wlAddr.s_addr){
		ForwardDecapMsg(&bsreq);
		return 0;
	}
	
	switch (bsreq.type) {
	case UNENCAP_ADD:
		AddMobile(eStatePtr, &bsreq);
		break;
		
	case UNENCAP_DIS:
		DisableMobile(eStatePtr, &bsreq);
		break;
		
	case UNENCAP_ENA:
	case UNENCAP_FORCE_ENA:
		EnableMobile(eStatePtr, &bsreq);
		break;
		
	case UNENCAP_DEL:
		DelMobile(eStatePtr, &bsreq);
		break;
	}
	
	return(0);
}

void main(int argc, char *argv[])
{
	int i, j, temp;
	int change;
	DecapdState eState;
	struct hostent *hp;
	char *tmpPtr;
	int rc;
	
	/* The argument ParseConfFile expects that any string pointers
	   passed as arguments have room for a strcpy() of a new value.
	   So we malloc() a new string and copy the old value to this
	   string. */
	tmpPtr=wiName;
	wiName=(char *) malloc(100*sizeof(char));
	strcpy(wiName, tmpPtr);

	tmpPtr=wlNames;
	wlNames=(char *) malloc(100*sizeof(char));
	strcpy(wlNames, tmpPtr);

	/* Use the values in the configuration file for the decap
           port, wired name, and wireless name. */
	ParseConfFile(confFileName, 3,
		      "decapPort", "%d", &decapPort,
		      "wiName", "%s", wiName,
		      "wlNames", "%[^\n]", wlNames);

	Opt_Parse (argc, argv, optionArray, Opt_Number(optionArray), 0);
	SetDebugLevel(debug);

	/* Initialize user-level encap state */
	if ((eState.wiAddr.s_addr = inet_addr(wiName)) == -1) {
		if ((hp = gethostbyname(wiName))!=0) {
			memcpy((char *)&eState.wiAddr, (char *)hp->h_addr,
			       sizeof(eState.wiAddr));
		} else 	{
			perror("wiAddr:");
			exit(-1);
		}
	}

	if ((eState.wlAddr.s_addr = inet_addr(wlNames)) == -1) {
		if ((hp = gethostbyname(wlNames))!=0) {
			memcpy((char *)&eState.wlAddr, (char *)hp->h_addr,
			       sizeof(eState.wlAddr));
		} else 	{
			perror("wlAddr:");
			exit(-1);
		}
	}	

	/* Initialize linked list of active connections */
	eState.connList.nextSock=NULL;

	/* Setup socket for controling kernel-level decap state */
	eState.ctrlFD = socket(AF_INET, SOCK_STREAM, 0);
	i = 1;
	j = sizeof(i);
	if (setsockopt(eState.ctrlFD, IPPROTO_TCP, TCP_NODELAY,(char *)&i,j) < 0)
		perror("Sock_Socket: Couldn't set tcp_nodelay option");
	
	/* Now erase all kernel-level state */
	ClearKernelState();
	
	/* Also reset when the program is killed for any reason */
	signal(SIGHUP, ClearKernelStateExit);
	signal(SIGINT, ClearKernelStateExit);
	signal(SIGQUIT, ClearKernelStateExit);
	signal(SIGTERM, ClearKernelStateExit);
	
	/* Setup socket for mobile-basestation communications */
#ifdef UDP
	eState.sdPtr = Sock_Socket(SOCK_UDP);
#else
	eState.sdPtr = Sock_Socket(SOCK_TCP);
#endif	 
	Sock_SetCompatible(eState.sdPtr, 1);
	Sock_SetBufferSize(eState.sdPtr, 16384);
	Sock_SetNoDelay(eState.sdPtr, 1);
	i = IPTOS_LOWDELAY;
	j = sizeof(i);
	if (setsockopt(Sock_GetFD(eState.sdPtr), IPPROTO_IP, IP_TOS, 
		       (char *)&i,j) < 0)
		perror("Sock_Socket: ip tos");
	
	if ((temp = Sock_Listen(eState.sdPtr, decapPort )) == SOCK_FAILURE) 
		exit(-1);
	
	NewFDSet(&eState);
	
	/* Setup socket for route changes */
	if ((eState.routes = socket(PF_ROUTE, SOCK_RAW, 0)) < 0) {
		perror("route socket");
		exit(1);
	}
	shutdown(eState.routes, 0); /* Don't want to read back our messages */
	
	/* Main loop: (1) read a msg (2) Process msg (3) sleep  */
	while (1) {
		change=0;
		dprintf(20,"just before select\n");
		if ((rc=select(FD_SETSIZE, &eState.fdvar,(fd_set *)0, (fd_set *)0, NULL))==-1){
			dprintf(10, "error in select()\n");
			continue;
		}

		if (FD_ISSET(Sock_GetFD(eState.sdPtr), &eState.fdvar)) {
			/* This is a new connection */
			SocketList *tempCdPtr=(SocketList *) malloc(sizeof(SocketList));

			dprintf(20,"in the if part\n");
			if ((tempCdPtr->sock = Sock_Accept(eState.sdPtr)) == (Socket *)SOCK_FAILURE){
				fprintf(stderr, "Couldn't accept() new connection, exiting..\n");
				exit(-1);
			}

			/* Add this connection to the linked list of connections */
			tempCdPtr->nextSock=eState.connList.nextSock;
			eState.connList.nextSock=tempCdPtr;
		} else {
		
			/* This is an existing connection */
			SocketList *tmpPtr, *prevPtr;
			fd_set fdvar;
			struct timeval zero;

			zero.tv_sec=0;
			zero.tv_usec=0;
			
			prevPtr=&(eState.connList);
			for (tmpPtr=eState.connList.nextSock; tmpPtr!=NULL; prevPtr=tmpPtr, tmpPtr=tmpPtr->nextSock) {
				FD_ZERO(&fdvar);
				FD_SET(Sock_GetFD(tmpPtr->sock), &fdvar);
				if (FD_ISSET(Sock_GetFD(tmpPtr->sock), &eState.fdvar) &&
				    (rc=select(FD_SETSIZE, &fdvar,(fd_set *)0, (fd_set *)0, &zero)==1)){
					
					if (ProcessDecapMsg(tmpPtr->sock, &eState) == -1) {
						/* Remove this entry from the linked list and free it */
						prevPtr->nextSock=tmpPtr->nextSock;

						/* Close control socket to mobile */
						Sock_Close(tmpPtr->sock);
						Sock_Free(tmpPtr->sock);
						free(tmpPtr);
					}
					change = 1;
				}
				if (change)
					goto nextLoop;
			}
		}
	  nextLoop:
		NewFDSet(&eState);
	}
	
	printf("hit key to end\n");
	getc(stdin);
	exit(0);
}







































































































