/*
 * Copyright (c) 1995, 1996, 1997
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* 
 * User-level mobile-side daemon for Daedalus Project's handoff software.
 * Original version written by Srinivasan Seshan (srini@watson.ibm.com)
 * modified heavily and released by Mark Stemm (stemm@cs.berkeley.edu)
 *
 * Daedalus group, Dept. of EECS, U.C. Berkeley.
 * Contact: stemm@cs.berkeley.edu, net@daedalus.cs.berkeley.edu
 */

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include <signal.h>
#include <unistd.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/in_var.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <kvm.h>

#include "ip_mobile.h"
#include "easysocklib.h"
#include "option.h"
#include "dbgprintlib.h"

#include "sample.h"
#include "wavelan.h"
#include "beacon.h"
#include "decap.h"
#include "newroute.h"
#include "vroute.h"

#include <advice_state.h>
#include <handoffctlr.h>
#include <conf.h>
#include <kread.h>

struct nlist nl[] = {
#define	N_IFNET	0
	{"_ifnet"},
#define	N_IFINDEX	1
	{"_if_index"},
#define	N_IN_IFADDR	2
	{"_in_ifaddr"},
};

kvm_t *kvmd;
char *memf = NULL;
char *nlistf = NULL;

#define NUM_UNENCAP_SEND 1

#define timerdiff(nowp, thenp)                                          \
(((nowp)->tv_sec - (thenp)->tv_sec) * 1000000 +                 \
 (nowp)->tv_usec - (thenp)->tv_usec)                    
	
int 	forceSend=1;
int     debug = 0;
int     burstThreshold = 10;
int     beaconPort = BEACON_PORT;
Port    decapPort = BSR_PORT;
int     beaconSwitchTimeout = 3000000;
int     analyzeStateInterval = 1000000;
char    *wiName = "chicago";
char    *multiName = "224.4.4.51";
char    *reverseTunnelAddr = "204.161.92.68";
char    *confFileName = "/etc/daedalus.conf";
char    *reqProto = "tcp";

Socket *adviceReq_sdPtr;	/* socket for receiving advice messages */
Socket *adviceRepl_sdPtr;	/* socket for sending advice replies back */
Socket *state_sdPtr;		/* socket for sending/receiving state messages */
Socket *beacon_sdPtr;		/* Socket for listening to beacons */

char *userAddr = "localhost";	/* address for communicating with the user
				 * controller */
int adviceReqPort = ADVICE_REQUEST_PORT; /* port for receiving advice
					  * requests */
int adviceReplPort = ADVICE_REPLY_PORT;	/* port for sending replies */

int stateReqPort = HANDOFFCTLR_REQUEST_PORT; /* port for receiving
					      * state requests */
int stateReplPort = HANDOFFCTLR_REPLY_PORT; /* port for sending replies */
int statePort = 4446;		
RouteState *sysStatePtr;	/* global pointer to the state information*/

Option  optionArray[] = {
	{OPT_INT, "debug", (char *) &debug,
	 "Turn on debugging (higher value -> more debugging)."},
	{OPT_INT, "beaconPort", (char *) &beaconPort,
	 "port for beacon messages"},
	{OPT_INT, "reverseTunnelAddr", (char *) &reverseTunnelAddr,
	 "Reverse Tunnel all outgoing packets through reverseTunnelAddr"},
	{OPT_INT, "burstThreshold", (char *) &burstThreshold,
	 "Number of packets in burst across 1 NI before switching to that NI"},
	{OPT_INT, "decapPort", (char *) &decapPort,
	 "port for decap messages"},
	{OPT_INT, "beaconSwitchTimeout", (char *) &beaconSwitchTimeout,
	 "Timeout before trying new base station/overlay (in usecs)"},
	{OPT_STRING, "confFileName", (char *) &confFileName,
	 "Configuration file for paramaters (default /etc/daedalus.conf)."},
	{OPT_STRING, "wiName", (char *) &wiName,
	 "hostname on wired net"},
	{OPT_STRING, "multiName", (char *) &multiName,
	 "multicast encapsulation addr"},
	{OPT_STRING, "rproto", (char *) &reqProto,
	 "UDP or TCP for requests to base stations"}
};

void SetIntervalTimer(int sec)
{
	struct itimerval timer;
	
	timer.it_value.tv_usec=0;
	timer.it_value.tv_sec=1;
	timer.it_interval.tv_usec=0;
	timer.it_interval.tv_sec=0;
	if (setitimer(ITIMER_REAL, &timer, NULL)<0){
		perror("Can't set AnalyzeState() timer");
	}
}

#define CancelStateTimer() SetIntervalTimer(0)
#define SetStateTimer() SetIntervalTimer(3)

int RefreshMultis(char *queryIfName)
{
	struct ifnet *ifp_addr;
	struct ifnet ifp;
	struct in_ifaddr *ia_addr;
	struct in_ifaddr ia;
	struct in_multi *inm_addr;
	struct in_multi inm;
	char ifName[10];
	static int numMulticastGroups=0;
	static Socket *multiGroups[100];
	int i;
	char shortQueryIfName[100];
	
	dprintf(10, "In RefreshMultis()\n");
	
	strncpy(shortQueryIfName, queryIfName, strlen(queryIfName)-1);
	shortQueryIfName[strlen(queryIfName)-1]='\0';
	
	/* Loop through the list of ifp structures until we find the interface that matches our interface */
	kread(nl[N_IFNET].n_value, (char *) &ifp_addr, sizeof(struct ifnet *), kvmd);
	while(ifp_addr!=NULL){
		kread((u_long) ifp_addr, (char *) &ifp, sizeof(struct ifnet), kvmd);
		kread((u_long) ifp.if_name, (char *) &ifName, 10, kvmd);
		if (strcmp(queryIfName, ifName)==0){
			break;
		}
		ifp_addr=ifp.if_next;
	} 
	
	if (ifp_addr==NULL){
		fprintf(stderr, "Couldn't find interface %s\n", queryIfName);
		return(-1);
	}
	
	/* Found the ifp structure, now find the in_ifaddr structure associated with this ifp structure */
	kread(nl[N_IN_IFADDR].n_value, (char *)&ia_addr, sizeof(struct in_ifaddr *), kvmd);
	while (ia_addr!=NULL) {
		kread((u_long) ia_addr, (char *) &ia, sizeof(struct in_ifaddr), kvmd);
		if (ia.ia_ifp==ifp_addr){
			break;
		}
		ia_addr=ia.ia_next;
	}
	
	if (ia_addr==NULL){
		fprintf(stderr, "Couldn't find ia structure for interface %s\n", queryIfName);
		return(-1);
	}
	
	/* Leave each multicast group that we were in before */
	for(i=0; i<numMulticastGroups; i++){
		Sock_Close(multiGroups[i]);
		multiGroups[i]=NULL;
	}
	
	numMulticastGroups=0;
	
	/* We found the ia structure, now find the list of multicast
           addresses for this ia and join each multicast group. */
	inm_addr=ia.ia_multiaddrs;
	while(inm_addr!=NULL){
		kread((u_long) inm_addr, (char *) &inm, sizeof(struct in_multi), kvmd);
		fprintf(stderr, "Group is %s\n", inet_ntoa(inm.inm_addr));
		
		multiGroups[numMulticastGroups] = Sock_Socket(SOCK_UDP_MULTI);
		Sock_SetCompatible(multiGroups[numMulticastGroups], 1);
		Sock_SetBufferSize(multiGroups[numMulticastGroups], 16384);
		if (Sock_Connect(multiGroups[numMulticastGroups], inet_ntoa(inm.inm_addr), 4000) == SOCK_FAILURE) {
			perror("Multicast Sock_Connect()");
			exit(-1);
		}
		
		numMulticastGroups++;
		inm_addr=inm.inm_next;
	}
	
	return(0);
}


/* From the string that describes the interface (wlp0, irp0, etc),
   determine what kind of interface it is */
int 
GetInterface(IFace * iPtr)
{
	int     ifNum = -1;
	
	if (iPtr->ifname[0] == 'w' && iPtr->ifname[1] == 'l') { /* Wavelan */
		iPtr->iftype = WAVELAN_TYPE;
		if (iPtr->ifname[2] == 'g' && iPtr->ifname[3] == 'p')
			ifNum = iPtr->ifname[4] - '0';	/* 2.4 GHz PCMCIA */
		else if (iPtr->ifname[2] == 'p')
			ifNum = iPtr->ifname[3] - '0';	/* 915 MHz PCMCIA */
		else
			ifNum = iPtr->ifname[2] - '0';	/* ISA */
	} else if (iPtr->ifname[0] == 'e' && iPtr->ifname[1] == 'f') { /* Ethernet */
		iPtr->iftype = ETHERNET_TYPE;
		if (iPtr->ifname[2] == 'p')
			ifNum = iPtr->ifname[3] - '0';	/* PCMCIA */
		else
			ifNum = iPtr->ifname[2] - '0';  /* ISA */
	} else if (iPtr->ifname[0] == 's' && iPtr->ifname[1] == 'l') { /* Metricom (using Slip) */
		iPtr->iftype = METRICOM_TYPE;
		ifNum = iPtr->ifname[2] - '0';
	} else if (iPtr->ifname[0] == 'i' && iPtr->ifname[1] == 'r') {	/* IR */
		iPtr->iftype = INFRARED_TYPE;
		if (iPtr->ifname[2] == 'p')
			ifNum = iPtr->ifname[3] - '0';	/* PCMCIA */
		else
			ifNum = iPtr->ifname[2] - '0';  /* ISA */
	} else if (iPtr->ifname[0]== 'p' && iPtr->ifname[1] == 'p' && iPtr->ifname[2] == 'p'){ /* Metricom (using PPP) */
		iPtr->iftype = METRICOM_TYPE;
		ifNum = iPtr->ifname[3] - '0';
	} else {
		/* Assign to "other" class (assumed to be worse than every other interface) */
		iPtr->iftype = OTHER_TYPE;
		ifNum = iPtr->ifname[3] - '0';
		dprintf(1, "unknowninterface %s\n", iPtr->ifname);
	}
	return ifNum;
}


/*
 * Send a message (request) to a base station.  This message instructs
 * the base station to make changes to its state with respect to the
 * sender of the request, the mobile host.
 */
int 
SendUnencapReq(BSRTypes type, PacketType pkttype, BSState * bsPtr, struct in_addr locAddr,
	       struct in_addr realAddr, struct in_addr multiAddr,
	       struct mobileip_ids * idsPtr)
{
	BSReq   bsreq;
	BS_States oldState;
	PacketType oldhdrOnly;
	int     i, j = sizeof(i);
	int forceSend = 0;
	Socket *sendToPtr;
	
	dprintf(20, "Entering SendUnencapReq()\n");
	
	/* If UNENCAP_FORCE_ENA is the type, then make sure to send out
	   the request anyway (for timing purposes) */
	if (type==UNENCAP_FORCE_ENA){
		type=UNENCAP_ENA;
		forceSend=1;
	}
	
	oldhdrOnly=bsPtr->hdrOnly;
	
	if (pkttype==HEADER){
		bsreq.hdrOnly=bsPtr->hdrOnly=1;
	} else {
		bsreq.hdrOnly=bsPtr->hdrOnly=0;
	}
	
	oldState = bsPtr->state;
	
	if (bsPtr->sdPtr == NULL) {
		if (!strcmp(reqProto, "udp")) {
			printf("using UDP\n");
			bsPtr->sdPtr = Sock_Socket(SOCK_UDP);
		} else
			bsPtr->sdPtr = Sock_Socket(SOCK_TCP);
		
		Sock_SetCompatible(bsPtr->sdPtr, 1);
		Sock_SetBufferSize(bsPtr->sdPtr, 16384);
		if (Sock_Connect(bsPtr->sdPtr, bsPtr->wlName,
				 decapPort) == SOCK_FAILURE)
			return (-1);
		Sock_SetNoDelay(bsPtr->sdPtr, 1);
		i = IPTOS_LOWDELAY;
		j = sizeof(i);
		if (setsockopt(Sock_GetFD(bsPtr->sdPtr), IPPROTO_IP, IP_TOS,
			       (char *) &i, j) < 0)
			perror("Sock_Socket: ip tos");
	}
	switch (type) {
	case UNENCAP_ADD:
		if (bsPtr->state != BS_FORWARDING)
			bsPtr->state = BS_BUFFERING;
		else
			return (0);
		break;
	case UNENCAP_DIS:
		if (bsPtr->state == BS_FORWARDING)
			bsPtr->state = BS_BUFFERING;
		else
			return (0);
		break;
	case UNENCAP_ENA:
		if (bsPtr->state == BS_BUFFERING || forceSend)
			bsPtr->state = BS_FORWARDING;
		else
			return (0);
		break;
	case UNENCAP_DEL:
		if (bsPtr->state != BS_INACTIVE)
			bsPtr->state = BS_INACTIVE;
		else
			return (0);
		break;
	default:
		printf("unknown BSRType %d\n", type);
		return (-1);
	}
	if (oldState == bsPtr->state && oldhdrOnly==bsPtr->hdrOnly && !forceSend)
		return (0);
	bsreq.type = type;
	bsreq.recipientAddr.s_addr = bsPtr->wlAddr.s_addr;
	bsreq.locAddr.s_addr = locAddr.s_addr;
	bsreq.realAddr.s_addr = realAddr.s_addr;
	bsreq.multiAddr.s_addr = multiAddr.s_addr;
	
	dprintf(20, "In SendUnencapReq(): forBS=%s: type=%d; locAddr=%x; realAddr=%x; multiAddr=%x hdrOnly=%d oldhdrOnly=%d state=%d oldState=%d\n",
		bsPtr->wlName, bsreq.type, bsreq.locAddr.s_addr,
		bsreq.realAddr.s_addr, bsreq.multiAddr.s_addr,
		bsreq.hdrOnly, oldhdrOnly, bsPtr->state, oldState);
	if (idsPtr != NULL)
		memcpy(&bsreq.ids, idsPtr, sizeof(bsreq.ids));
	else
		memset(&bsreq.ids, 0, sizeof(bsreq.ids));
	
	/* Always send unencap messages to the current base
           station. The current base station will forward the unencap
           message to the correct base station if necessary */
/*	if (sysStatePtr->curiPtr==NULL ||
	    sysStatePtr->curiPtr->curBSPtr==NULL){*/
	if (1){
		sendToPtr=bsPtr->sdPtr;
	} else {
		sendToPtr=sysStatePtr->curiPtr->curBSPtr->sdPtr;
	}
	
	for (i = 0; i < NUM_UNENCAP_SEND; i++)
		if (Sock_Send(sendToPtr,
			      (char *) &bsreq, sizeof(bsreq)) == SOCK_FAILURE) {
			sysStatePtr->curiPtr->curBSPtr->state = BS_DEAD;
			return (-1);
		}
	return (0);
}

/* This removes all base stations that haven't been heard from for a while */
void
ClearOld(RouteState * sysStatePtr, IFace * iPtr, struct timeval * curTimePtr)
{
	int     i, j, lastSamp;
	BSState *temp;
	
	dprintf(70, "In ClearOld()...\n");
	
	if (iPtr==NULL){
		return;
	}
	
	for (i = 0; i < iPtr->numBS; i++) {
		lastSamp = iPtr->BSPtr[i]->lastSamp;
		if ((curTimePtr->tv_sec > (iPtr->BSPtr[i]->samples[lastSamp].time.tv_sec + BEACON_TIMEOUT)) ||
		    iPtr->BSPtr[i]->state == BS_DEAD) {
			/* old entry */
			dprintf(5, "deling bs %s from %d\n",
				iPtr->BSPtr[i]->wlName, i);
			SendUnencapReq(UNENCAP_DEL, PACKET, iPtr->BSPtr[i], iPtr->locAddr,
					   sysStatePtr->realAddr,
					   sysStatePtr->multiAddr, NULL);
			DelBeacon(iPtr->ifs, iPtr->BSPtr[i]->wlName, iPtr->ifname);
			Sock_Close(iPtr->BSPtr[i]->sdPtr);
			Sock_Free(iPtr->BSPtr[i]->sdPtr);
			iPtr->BSPtr[i]->wlAddr.s_addr = 0;	/* mark empty */
			
			/* remove gap in array */
			temp = iPtr->BSPtr[i];
			for (j = i + 1; j < iPtr->numBS; j++)
				iPtr->BSPtr[j - 1] = iPtr->BSPtr[j];
			iPtr->BSPtr[iPtr->numBS--] = temp;
			if (iPtr->curBSPtr == temp)
				iPtr->curBSPtr = NULL;
			iPtr->numBS--;
		}
	}
}

/* This function compares two overlay networks to determine which is
   "better". It takes into account the static ordering of the overlay
   hierarchy as well as the recent presence/absence of beacon packets
   and burst patterns */
int CompareOverlays(const void *if1v, const void *if2v)
{
	IFace *if1, *if2, *higherIf, *lowerIf;
	int decision=0;
	struct timeval timeSinceLastBeacon, timeSincePrevLastBeacon, curTime;
	int size, measured;
	struct mobilemh_state state;
	
	if1=*(IFace **) if1v;
	if2=*(IFace **) if2v;
	
	/* By default, if an overlay is lower in the hierarchy it
	  ranks better. If the 2 overlays are adjacent, two things
	  can override this decision:
	  1. The higher overlay has seen a burst, while the lower hasn't.
	  2. The lower hasn't heard a beacon for the beacon timeout period.
	  If 1 or 2 is true, then the higher overlay is better. */
	if (if1->iftype>if2->iftype){
		higherIf=if1;
		lowerIf=if2;
		decision=1;
	} else {
		higherIf=if2;
		lowerIf=if1;
		decision=-1;
	}

	/* If the higher overlay has no active base stations, the
           lower overlay is better. If the lower overlay has no active
           base stations, the higher overlay is better */
	if (higherIf->numBS==0){
		/* The lower interface is better, the default decision was right. */
		return(decision);
	}

	if (lowerIf->numBS==0){
		/* The higher interface is better, change the decision */
		if (if1->iftype>if2->iftype){
			decision=-1;
		} else {
			decision=1;
		}
		goto end;
	}

	gettimeofday(&curTime, NULL);
	
	/* Find out how many seconds have gone by since the */
	/* last beacon has been received on this interface */
	timeSinceLastBeacon.tv_sec=curTime.tv_sec-lowerIf->lastBeaconTime.tv_sec;
	timeSinceLastBeacon.tv_usec=curTime.tv_usec-lowerIf->lastBeaconTime.tv_usec;
	if (timeSinceLastBeacon.tv_usec<0){
		timeSinceLastBeacon.tv_sec--;
		timeSinceLastBeacon.tv_usec+=1000000;
	}
				
	/* Also find out how many seconds have gone by
	   since the second-to-last beacon has been
	   received on this interface */
	timeSincePrevLastBeacon.tv_sec=curTime.tv_sec-lowerIf->prevLastBeaconTime.tv_sec;
	timeSincePrevLastBeacon.tv_usec=curTime.tv_usec-lowerIf->prevLastBeaconTime.tv_usec;
	if (timeSincePrevLastBeacon.tv_usec<0){
		timeSincePrevLastBeacon.tv_sec--;
		timeSincePrevLastBeacon.tv_usec+=1000000;
	}
	
	/* Change our decision if: 1. higher overlay has seen a burst,
           or 2. Lower overlay hasn't seen beacons for more than the timeout period */
	measured=timeSinceLastBeacon.tv_sec*1000000+timeSinceLastBeacon.tv_usec;
	dprintf(70, "measured time=%d usec, timeouit=%d usec\n", measured, beaconSwitchTimeout);
	if (timeSinceLastBeacon.tv_sec*1000000+timeSinceLastBeacon.tv_usec >= beaconSwitchTimeout){
		/* The higher interface is better, change the decision */
		if (if1->iftype>if2->iftype){
			decision=-1;
		} else {
			decision=1;
		}		
	}
  end:
	if (decision==1){
		dprintf(70, "iface %s is lower (better) than %s\n", if2->ifname, if1->ifname);
	} else {
		dprintf(70, "iface %s is lower (better) than %s\n", if1->ifname, if2->ifname);
	}
	return(decision);
}

/* This function compares two base stations to determine which is
   "better". This depends on the signal strength of the beacons. */
int CompareBaseStations(const void *bs1v, const void *bs2v)
{
	BSState *bs1, *bs2;
	struct timeval curTime;
	
	bs1 = *(BSState **) bs1v;
	bs2 = *(BSState **) bs2v;
	
	/* If the time of the last sample for this BS is more than
           beaconTimeout seconds old, then make the signal strength level = 0 */
	gettimeofday(&curTime, NULL);
	if (timerdiff(&(bs1->samples[bs1->lastSamp].time), &(curTime))>=5000000){
		bs1->samples[bs1->lastSamp].siglvl=0;
	}
	
	if (timerdiff(&(bs2->samples[bs2->lastSamp].time), &(curTime))>=5000000){
		bs2->samples[bs2->lastSamp].siglvl=0;
	}
	
	if ((bs1->samples[bs1->lastSamp].siglvl >
	     bs2->samples[bs2->lastSamp].siglvl)){
		return(-1);
	} else if ((bs1->samples[bs1->lastSamp].siglvl <
	     bs2->samples[bs2->lastSamp].siglvl)){
		return(1);
	} else {
		return(0);
	}
}

/* This fuction telle the Mobile IP layer to reverse Tunnel/not reverse Tunnel all outgoing packets through the home agent */
void ReverseTunnel(char *reverseTunnelAddr)
{
	struct ip_mobreq req;
	int fd, j, option;
	struct timeval start;
	char errmsg[2000];
	
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd<0){
		perror("While getting socket to tell NI");
	}
	
	j=sizeof(req);
	option = IP_TUNNEL_ADDR;
	inet_aton(reverseTunnelAddr, &req.imr_ins_addr);
	gettimeofday(&start, NULL);
	
	if (option)
		if (setsockopt(fd,IPPROTO_IP, option,(char *)&req,j) < 0)
			perror(errmsg);
}
	

/* This function tells the Mobile IP layer to allow/disallow packets via this interface */
void TellNI(BSRTypes type, IFace *iPtr)
{
	struct ip_mobreq req;
	int fd, j, option;
	struct timeval start;
	char errmsg[2000];
	char command[200];
	char locAddr[40];
	char netMask[40];
	int i;
	struct in_addr newAddr;
	
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd<0){
		perror("While getting socket to tell NI");
	}
	
	j=sizeof(req);
	if (type==UNENCAP_ENA){
		option = IP_IF_ON;
		/* Don't copy the 0 at the end */
		strncpy(req.blockedIf, iPtr->ifname, strlen(iPtr->ifname)-1);
		req.blockedIf[strlen(iPtr->ifname)-1]=NULL;
		gettimeofday(&start, NULL);
		dprintf(1, "Turned on if at %d secs, %d usecs\n", start.tv_sec, start.tv_usec);

		/* Also add a route for the wireless net for this interface */
		/* The routes for the base stations on this net will be added */
		/* By arp. We must manually add routes for all "foreign" base
		   stations associated with this net.*/
		strcpy(locAddr, inet_ntoa(iPtr->locAddr));
		sprintf(netMask, "0x%x", ntohl(iPtr->netMask.s_addr));
		sprintf(command, "ifconfig %s %s netmask %s", iPtr->ifname, locAddr, netMask);
		dprintf(80, "In TellNI(): Command=\"%s\"", command);
		system(command);
		for(i=0; i<iPtr->numBS; i++){
			if (iPtr->BSPtr[i]->foreignBS){
/*				AddHostRoute(iPtr->BSPtr[i]->wlName, iPtr->ifname, sysStatePtr->routes);*/
			}
		}
	} else if (type==UNENCAP_DIS){
		option = IP_IF_OFF;
		/* Don't copy the 0 at the end */
		strncpy(req.blockedIf, iPtr->ifname, strlen(iPtr->ifname)-1);
		req.blockedIf[strlen(iPtr->ifname)-1]=NULL;

		/* Also delete the route for the network and all base stations on the interface */

		newAddr.s_addr=iPtr->netAddr.s_addr;
		sprintf(command, "route delete %s", inet_ntoa(newAddr));
		dprintf(80, "In TellNI(): Command=\"%s\"", command);
		system(command);
		for(i=0; i<iPtr->numBS; i++){
			sprintf(command, "route delete %s", iPtr->BSPtr[i]->wlName);
			dprintf(80, "In TellNI(): Command=\"%s\"", command);
			system(command);
		}
	}
	
	sprintf(errmsg, "Turning ni %s on/off using socket %d", iPtr->ifname, fd);
	
	if (option)
		if (setsockopt(fd,IPPROTO_IP, option,(char *)&req,j) < 0)
			perror(errmsg);
}

void GetIDs(struct mobileip_ids *ids, BSState *bsPtr)
{
	int j;
	j=sizeof(*ids);
	if (getsockopt(Sock_GetFD(bsPtr->sdPtr), IPPROTO_IP, IP_GET_IDS, (char *) ids, &j) < 0)
		perror("Sock_Socket: ip tos");
	dprintf(25, "got ids: %d: %d %d %d\n", ids->lst_pkt, ids->ip_id[0], ids->ip_id[1], ids->ip_id[2]);
	
}

/* This function tells the specified base station to start
   decapulating and sets the default route of the mobile device to
   that base station. It also informs the handoff controller of the
   new base station and interface. */
void SwitchToBS(RouteState *sysStatePtr, BSState *newBSPtr, IFace *newiPtr)
{
	struct mobileip_ids ids;
	
	/* Also, if using low latency, tell the kernel to forward packets again */
	TellNI(UNENCAP_ENA, newiPtr);

	dprintf(10, "In SwitchToBS()\n");
	GetIDs(&ids, newBSPtr);
	dprintf(10, "enabled %s\n", newBSPtr->wlName);
	if (SendUnencapReq(UNENCAP_ENA, PACKET, newBSPtr, newiPtr->locAddr, sysStatePtr->realAddr,
			   sysStatePtr->multiAddr, &ids) == -1) {
		dprintf(5, "SwitchToBS(): SendUnencapReq() UNENCAP_ENA failed\n");
		return;
	}
	
	SetDefaultRoute(newBSPtr->wlName, sysStatePtr->routes);

	/* To make sure that multicast packets will still be recieved,
	   we must rejoin all of the multicast groups for this mobile
	   on the new base station. */
/*	 RefreshMultis(newiPtr->ifname);*/
}

/*
 * Do handoff from curiPtr->curBSPtr to newBSPtr.
 */
void
DoHandoff(RouteState * sysStatePtr, IFace * oldiPtr, IFace * newiPtr, BSState *oldBSPtr, BSState *newBSPtr)
{
	struct timeval now;
	
	gettimeofday(&now, NULL);
	
	dprintf(5, "handoff to %s, at %d sec %d usec \n", newBSPtr->wlName, now.tv_sec, now.tv_usec);

	sysStatePtr->curiPtr=newiPtr;
	newiPtr->curBSPtr=newBSPtr;
	
	/* Switch to the New Base Station first */
	SwitchToBS(sysStatePtr, newBSPtr, newiPtr);
	
	/* If we are switching interfaces, trun off the old interface */
	if (oldiPtr!=newiPtr && oldiPtr!=NULL){
		TellNI(UNENCAP_DIS, oldiPtr);
	}

	if (oldBSPtr!=NULL){
		dprintf(10, "disabled %s\n", oldBSPtr->wlName);
		if (SendUnencapReq(UNENCAP_DIS, PACKET, oldBSPtr, oldiPtr->locAddr, sysStatePtr->realAddr, sysStatePtr->multiAddr, NULL) == -1) {
			dprintf(5, "DoHandoff(): SendUnencapReq() UNENCAP_DIS failed\n");
			return;
		}
	}
}

/* This is a Wrapper around AnalyzeState that has no arguments. It is
   called when the SIGALRM signal is caught (meaning that the mobile
   has moved out of range of the current overlay */
void AnalyzeStateWrapperSIGALRM()
{
	
	dprintf(10, "Caught SIGALRM signal\n");
	
	AnalyzeState(sysStatePtr);

}		  

/* This is a Wrapper around AnalyzeState that has no arguments. It is
   called when the SIGUSR1 signal is caught (meaning that the mobile
   has moved out of range of the current overlay */
void AnalyzeStateWrapperSIGUSR1()
{
	struct timeval start;
	
	gettimeofday(&start, NULL);
	
	dprintf(1, "Caught SIGUSR1 signal at %d sec, %d usec\n", start.tv_sec, start.tv_usec);
	
	AnalyzeState(sysStatePtr);
}		  

void EnableDisableBuffering(IFace *iPtr)
{
	int i, worstBS = iPtr->numBS - 1;

	/* Enable buffering on all base stations that are heard */
	iPtr->desBufing = iPtr->numBS;
	
	/* enable buffering on determined BS's */
	for (i = 0; i < iPtr->desBufing; i++) {
		if (iPtr->BSPtr[i]->state == BS_INACTIVE) {
			if (SendUnencapReq(UNENCAP_ADD, PACKET, iPtr->BSPtr[i], 
					   iPtr->locAddr,
					   sysStatePtr->realAddr,
					   sysStatePtr->multiAddr, NULL) == -1) {
				continue;
			}
		}
	}
	
	/* disable buffering on determined BS's */
	for (i = iPtr->desBufing; i < worstBS + 1; i++) {
		if (iPtr->BSPtr[i]->state == BS_BUFFERING) {
			if (SendUnencapReq(UNENCAP_DEL, PACKET, iPtr->BSPtr[i], 
					   iPtr->locAddr,
					   sysStatePtr->realAddr,
					   sysStatePtr->multiAddr, NULL) == -1) {
				continue;
			}
		}
	}
	
}

/* Analyzes Beacon information to find the base station to use. A
state diagram that shows how this function chooses the base station
to use can be found at http://daedalus.cs.berkeley.edu/vhandoff/design.html. */

void AnalyzeState(RouteState * sysStatePtr)
{
	int     i;
	IFace  *newiPtr=NULL;	/* pointer to the new interface */
	IFace *curiPtr;		/* temp pointer for loop indexing */
	BSState *newBestBS;	/* pointer to new "best" BS */
	BSState *curBSPtr;	/* pointer to last best BS of best overlay */
	static BSState *candidateBS=NULL; /* a BS to switch to if it stays the best for long enough */
	int validBS;		/* any base station around? */
	static HandoffState handoffState= INACTIVE;
	static struct timeval waitStartTime;
	struct timeval now;
	char name[100];

	/* First remember what the current interface and base stations are */
	curiPtr=sysStatePtr->curiPtr;
	if (curiPtr==NULL){
		curBSPtr=NULL;
	} else {
		curBSPtr=curiPtr->curBSPtr;
	}

	/* Sort the base stations for each overlay */
	validBS=0;
	for (i=0; i<sysStatePtr->numIfaces; i++) {
		if (sysStatePtr->iface[i] && sysStatePtr->iface[i]->numBS>0) {
			validBS=1;
			sysStatePtr->iface[i]->oldBestBS=sysStatePtr->iface[i]->BSPtr[0];

			/* Sort the interfaces based on signal strength */
			qsort((void *) sysStatePtr->iface[i]->BSPtr, sysStatePtr->iface[i]->numBS,
			      sizeof(BSState *), CompareBaseStations);
		}
	}

	/* If no base stations, don't do anything */
	if (!validBS){
		handoffState=INACTIVE;
		return;
	}
				
	/* Now choose the best overlay. The best overlay is the
           "lowest" one who has been heard from recently. */
	qsort((void *) sysStatePtr->iface, sysStatePtr->numIfaces, sizeof(IFace *), CompareOverlays);

	/* The best base station of the best overlay is the one we
           want to switch to */
	newiPtr=sysStatePtr->iface[0];
	newBestBS=newiPtr->BSPtr[0];

	if (curBSPtr==NULL){
		strcpy(name, "Null");
	} else {
		strcpy(name, curBSPtr->wlName);
	}
	
	dprintf(5, "AState():curBS %s, best bs %s %s (%d times)\n", name, sysStatePtr->iface[0]->ifname, newBestBS->wlName, newBestBS->numBest);


	/* If in INACTIVE state, just switch to best BS */
	if (handoffState==INACTIVE){
		handoffState=NORMAL;
		dprintf(1, "Original handoff to %s %s\n", newiPtr->ifname, newBestBS->wlName);
		DoHandoff(sysStatePtr, NULL, newiPtr, NULL, newBestBS);
		return;
	}

	/* If best BS is same as current BS, go back to NORMAL state. */
	if (newBestBS==curBSPtr){
		handoffState=NORMAL;
		return;
	}

	/* If in NORMAL state and best BS is different from current
           BS, move to WAITING state and remember the time. */
	if (handoffState==NORMAL && newBestBS!=curBSPtr){
		handoffState=WAITING;
		gettimeofday(&waitStartTime, NULL);
		candidateBS=newBestBS;
		return;
	}

	/* If in WAITING state and best BS is different from candidate
	BS, remember the time and change the candidate BS */
	if (handoffState==WAITING && newBestBS!=candidateBS){
		handoffState=WAITING;
		gettimeofday(&waitStartTime, NULL);
		candidateBS=newBestBS;
		return;
	}
	
	/* We don't want to immediately jump to a new base station
	once it is the best. A base station must be the best for at
	least beaconSwitchTimeout seconds before moving to it. */

	/* If in WAITING state and best BS is same is candidate BS,
	but has not yet been beaconSwitchTime usecs since waitStartTime,
	keep waiting */
	gettimeofday(&now, NULL);
	if (handoffState==WAITING && timerdiff(&now, &waitStartTime)<=beaconSwitchTimeout){
		return;
	}

	/* If in WAITING state and best BS is same is candidate BS,
	and it has been beaconSwitchTime usecs since waitStartTime,
	handoff to BS. */
	if (handoffState==WAITING && timerdiff(&now, &waitStartTime)>=beaconSwitchTimeout){
		DoHandoff(sysStatePtr, curiPtr, newiPtr, curBSPtr, newiPtr->oldBestBS);
		handoffState=NORMAL;
	}
}


ResetDontTouchAddr()
{
	struct ip_mobreq req;
	int fd, j, option;
	char errmsg[2000];
	
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd<0){
		perror("While getting socket to tell NI");
	}
	
	j=sizeof(req);
	option = IP_CLR_DT_ADDR;

	if (option)
		if (setsockopt(fd,IPPROTO_IP, option,(char *)&req,j) < 0)
			perror(errmsg);
}

AddDontTouchAddr(struct in_addr addr)
{
	struct ip_mobreq req;
	int fd, j, option;
	char errmsg[2000];
	
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd<0){
		perror("While getting socket to tell NI");
	}
	
	j=sizeof(req);
	option = IP_ADD_DT_ADDR;
	req.imr_ins_addr.s_addr=addr.s_addr;
	
	if (option)
		if (setsockopt(fd,IPPROTO_IP, option,(char *)&req,j) < 0)
			perror(errmsg);
}


void AddNewBaseStation(BSState *bsPtr, IFace *iPtr, BeaconMsg *bmsg, RouteState *sysStatePtr)
{
	struct sockaddr_in localtemp;
	int t1, t2;


	memset(bsPtr, 0, sizeof(BSState));
	bsPtr->wlAddr.s_addr = bmsg->data.beacon.wlAddr.s_addr;
	bsPtr->wiAddr.s_addr = bmsg->data.beacon.wiAddr.s_addr;
	bsPtr->bAddr.s_addr = bmsg->data.beacon.bAddr.s_addr;
	strcpy(bsPtr->wlName, inet_ntoa(bmsg->data.beacon.wlAddr));
	strcpy(bsPtr->wiName, inet_ntoa(bmsg->data.beacon.wiAddr));
	++iPtr->numBS;

	/* Make sure outgoing packets to this BS aren't modified */
	AddDontTouchAddr(bsPtr->wlAddr);

	if ((bsPtr->bAddr.s_addr & iPtr->netMask.s_addr) != (bsPtr->bAddr.s_addr & iPtr->netMask.s_addr)){
		/* If the beacon address and the base station's wireless
		   address aren't on the same subnet, then this base station
		   is a "distant" base station advertising for a local wireless
		   net (for example, a Metricom or CDPD network). In these cases,
		   we must manually add routes for these base stations. */
		bsPtr->foreignBS=1;

		/* Add a host route for the foreign BS via the interface for this address */
/*		AddHostRoute(bsPtr->wlName, iPtr->ifname, sysStatePtr->routes);*/
	} else {
		bsPtr->foreignBS=0;
	}

	dprintf(10, "adding new bs %s\n", bsPtr->wlName);
	
	if (!strcmp(reqProto, "udp")) {
		printf("using UDP\n");
		bsPtr->sdPtr = Sock_Socket(SOCK_UDP);
	} else
		bsPtr->sdPtr = Sock_Socket(SOCK_TCP);

	localtemp.sin_addr.s_addr = iPtr->locAddr.s_addr;
	localtemp.sin_family = AF_INET;
	localtemp.sin_port = 0;
	bind(bsPtr->sdPtr->fd, (struct sockaddr *) & localtemp,
	     sizeof(localtemp));
	Sock_SetCompatible(bsPtr->sdPtr, 1);
	Sock_SetBufferSize(bsPtr->sdPtr, 16384);
	if (Sock_Connect(bsPtr->sdPtr, bsPtr->wlName, decapPort) == SOCK_FAILURE) {
		dprintf(5, "error adding new bs %s\n",
			bsPtr->wlName);
		bsPtr->state = BS_DEAD;
		return;
	} else {
		Sock_SetNoDelay(bsPtr->sdPtr, 1);
		t1 = IPTOS_LOWDELAY;
		t2 = sizeof(t1);
		if (setsockopt(Sock_GetFD(bsPtr->sdPtr),
			       IPPROTO_IP, IP_TOS,
			       (char *) &t1, t2) < 0)
			perror("Sock_Socket: ip tos");
		dprintf(5, "added new bs %s\n",
			bsPtr->wlName);
	}

	
	dprintf(5, "added new bs %s\n",
		bsPtr->wlName);

	if (iPtr->iftype == WAVELAN_TYPE){	/* currently works only
						 * for WaveLAN. TODO:
						 * extend to Metricom  */
		AddBeacon(iPtr->ifs, inet_ntoa(bmsg->data.beacon.wlAddr),iPtr->ifname);
	}
			
	/* default action: set the
	 * basestation in buffering mode */
	if (bsPtr->state == BS_INACTIVE) {
		if (SendUnencapReq(UNENCAP_ADD, PACKET, bsPtr,
				   iPtr->locAddr,
				   sysStatePtr->realAddr,
				   sysStatePtr->multiAddr, NULL) == -1) {
			
			dprintf(5, "SendUnencapReq(UNENCAP_ADD) failed\n");
		}
	}
	
}

void HandleBeaconPkt(BeaconMsg *bmsg, struct timeval *curTime, IFace *iPtr, RouteState *sysStatePtr)
{
	BSState *bsPtr;
	WLSamp  powerSamp;
	int temp, i;

	powerSamp.addr.s_addr = bmsg->data.beacon.wlAddr.s_addr;
	powerSamp.time.tv_sec = curTime->tv_sec;
	powerSamp.time.tv_usec = curTime->tv_usec;
	/* This needs to be changed for non-Wavelan
	 * interfaces. Get the power measure - if
	 * this a new BS add to list of beacons in
	 * kernel and wait for next packet. */
	if (iPtr->iftype == WAVELAN_TYPE) {
		if (!GetSample(iPtr->sampdPtr, iPtr->ifNum, &powerSamp)) {
			char    tempstr[16];
			strcpy(tempstr, inet_ntoa(bmsg->data.beacon.wlAddr));
			AddBeacon(iPtr->ifs, tempstr, iPtr->ifname);
			return;
		}
	} else {	/* non-Wavelan */
		powerSamp.siglvl = 63;	/* pretty random
					 * choices! */
		powerSamp.sillvl = 33;
		powerSamp.sigqual = 33;
	}

	dprintf(50, "Beacon has level %d\n", powerSamp.siglvl);
	
	/* Search for the BS in ("best") iPtr state. */
	temp = -1;
	for (i = 0; i < iPtr->numBS; i++){
		if (bmsg->data.beacon.wlAddr.s_addr == iPtr->BSPtr[i]->wlAddr.s_addr){
			break;
		}
	}
	
	/* After search for interfaces. bsPtr will be valid or a pointer to an unused slot */
	bsPtr = iPtr->BSPtr[i];
	
	if (bmsg->data.beacon.wlAddr.s_addr == bsPtr->wlAddr.s_addr){
		/* found */
		bsPtr->lastSamp =
			(bsPtr->lastSamp + 1) % MAX_SAMPLES;
	}
	else {
		
		/* not found ==> new BS */
		if (iPtr->numBS == MAX_BS) {
			dprintf(5, "vrouted: beacon table full\n");
			return;
		}

		/* The new Base Station must have a minimum signal strength before we try to connect to it */

		if (powerSamp.siglvl>10){
			AddNewBaseStation(bsPtr, iPtr, bmsg, sysStatePtr);
		}
	}
	
	/* update the stats for the entry  */
	dprintf(20, "updating bs %s entry %d level %d\n", bsPtr->wlName, bsPtr->lastSamp, powerSamp.siglvl);
	
	memcpy((void *) &(bsPtr->samples[bsPtr->lastSamp]),
	       (void *) &powerSamp, sizeof(powerSamp));
	
	/* TODO: extend this to "delete" interfaces
	 * (and the corresponding basestations that
	 * have been inactive for a while */
	/* delete any old base station entries */
	/* Loop over all interfaces */
	for (i = 0; i < sysStatePtr->numIfaces; i++){
		ClearOld(sysStatePtr, sysStatePtr->iface[i], curTime);
	}

}

void ProcessBeaconPacket(BeaconMsg *bmsg, int temp)
{
	static BeaconMsg beacons[MAX_INTERFACES];
	IFace *iPtr;
	int i;
	struct timeval curTime;
	char *sender;
	struct sockaddr_in sender_addr;
	
	/* If we are here, the packet must have been a beacon packet */
	/* Determine the address of the sender and the which of our
           interfaces that the packet was destined for. */
	
	sender=inet_ntoa(beacon_sdPtr->lastRSin.sin_addr);
	sender_addr.sin_addr.s_addr = ntohl(beacon_sdPtr->lastRSin.sin_addr.s_addr);
	iPtr = NULL;
	for (i = 0; i < sysStatePtr->numIfaces; i++) {
		iPtr = sysStatePtr->iface[i];

		/* The beacon address must be for one of the interfaces we have defined. */
		if ((bmsg->data.beacon.bAddr.s_addr & iPtr->netMask.s_addr) == iPtr->netAddr.s_addr){
			break; /* found the matching interface */
		}
	}
	
	if (i >= sysStatePtr->numIfaces) {	/* received a beacon on an
						 * unknown interface; just
						 * ignore */
		dprintf(20, "Received beacon on unknown interface; ignoring\n");
		return;
	}
	
	/* Update the time at which the last beacon was heard for this interface */
	iPtr->prevLastBeaconTime.tv_sec=iPtr->lastBeaconTime.tv_sec;
	iPtr->prevLastBeaconTime.tv_usec=iPtr->lastBeaconTime.tv_usec;
	gettimeofday(&iPtr->lastBeaconTime, NULL);
	bcopy((char *) bmsg, (char *) &(beacons[i]), sizeof(BeaconMsg));
	if (iPtr->active == TRUE && temp < 0) {
		if (temp == SOCK_FD_CLOSED)
			dprintf(10, "remote iface %s closed\n", iPtr->ifname);
		if (temp == SOCK_FAILURE)
			dprintf(10, "%s error in recv\n", iPtr->ifname);
		iPtr->active = FALSE;
		--sysStatePtr->numActiveIfaces;
		return;
	}
	if (iPtr->active == FALSE && temp > 0) {	/* First beacon for this interface */
		iPtr->active = TRUE;
		++sysStatePtr->numActiveIfaces;
	}
	
	if (iPtr == NULL) {
		dprintf(20, "numActiveIfaces %d\n", sysStatePtr->numActiveIfaces);
		dprintf(1, "no active interfaces\n");
		exit(1);
	}
	gettimeofday(&curTime, NULL);	/* approximate same time for
					 * all ifs */
	
	switch (bmsg->type) {
	case BEACON:
		HandleBeaconPkt(bmsg, &curTime, iPtr, sysStatePtr);
		break;
		
	default:
		fprintf(stderr,"Can only handle beacon packets\n");
	}	 
	
}

void ListenForBeacons(RouteState *sysStatePtr)
{
	fd_set fdvar;
	char    advicemsg[1000];/* maximum length of an advice message */
	char    statemsg[1000];/* maximum length of a state message */	 
	int temp;
	BeaconMsg bmsg;
	static struct timeval now, last;
	static int first=1;

	if (first){
		gettimeofday(&last, NULL);
		first=0;
	}

	FD_ZERO(&fdvar);
	FD_SET(beacon_sdPtr->fd, &fdvar);
	FD_SET(adviceReq_sdPtr->fd, &fdvar);
/*	FD_SET(state_sdPtr->fd, &fdvar);*/
	if (select(FD_SETSIZE, &fdvar, NULL, NULL, NULL) < 0) {
		perror("select()");
		return;
	}

	if (FD_ISSET(adviceReq_sdPtr->fd, &fdvar)) {	/* received a advice message */
		if ((temp = Sock_Recv(adviceReq_sdPtr, advicemsg, 1000)) < 0)
			perror("Sock_Recv(adviceReq_sdPtr)");
		else{
/*			ProcessAdviceRequest(advicemsg, sysStatePtr, sysStatePtr->curiPtr);*/
		}
	}
#if 0	
	if (FD_ISSET(state_sdPtr->fd, &fdvar)) {	/* received a state message */
		if ((temp = Sock_Recv(state_sdPtr, statemsg, 1000)) < 0)
			perror("Sock_Recv(state_sdPtr)");
		else
			ProcessStateRequest(statemsg, sysStatePtr, sysStatePtr->curiPtr);
	}
#endif	
	if (FD_ISSET(beacon_sdPtr->fd, &fdvar)) {	/* received a beacon packet */
		if ((temp = Sock_Recv(beacon_sdPtr, (char *) &bmsg, sizeof(BeaconMsg))) < 0)
			return;
		else{
			ProcessBeaconPacket(&bmsg, temp);
		}
		
	}

	/* If it's been longer than analyzeStateInterval usecs, AnalyzeState */
	gettimeofday(&now, NULL);

	if (timerdiff(&now, &last)>=analyzeStateInterval){
		AnalyzeState(sysStatePtr);
		last.tv_sec=now.tv_sec;
		last.tv_usec=now.tv_usec;		
	}
}

void CleanupStateExit()
{
	int i, j;
	IFace *iPtr;
	
	/* Delete all BS state */
	for (i = 0; i < sysStatePtr->numIfaces; i++){
		iPtr=sysStatePtr->iface[i];
		for (j = 0; j < iPtr->numBS; j++) {
			dprintf(5, "deling bs %s from %d\n",
				iPtr->BSPtr[j]->wlName, j);
			SendUnencapReq(UNENCAP_DEL, PACKET, iPtr->BSPtr[j], iPtr->locAddr,
				       sysStatePtr->realAddr,
				       sysStatePtr->multiAddr, NULL);
			DelBeacon(iPtr->ifs, iPtr->BSPtr[j]->wlName, iPtr->ifname);
			Sock_Close(iPtr->BSPtr[j]->sdPtr);
			Sock_Free(iPtr->BSPtr[j]->sdPtr);
		}
	}	
	
	/* Close all sockets */
/*	Sock_Close(state_sdPtr);*/

	exit(0);
}

void InitSockets()
{
	int temp;
	
	/* create and set up one socket for all interfaces to listen to
	 * beacons */
	beacon_sdPtr = Sock_Socket(SOCK_UDP);	/* for beacons */
	Sock_SetCompatible(beacon_sdPtr, 1);
	Sock_SetBufferSize(beacon_sdPtr, 16384);
	Sock_SetAcceptFlag(beacon_sdPtr, ntohl(INADDR_ANY));
	if ((temp = Sock_Listen(beacon_sdPtr, beaconPort)) == SOCK_FAILURE)
		perror("sock_listen()");
	
	/* create and set up a socket to receive advice messages from whoever */
	adviceReq_sdPtr = Sock_Socket(SOCK_UDP);
	Sock_SetCompatible(adviceReq_sdPtr, 1);
	Sock_SetBufferSize(adviceReq_sdPtr, 16384);
	Sock_SetAcceptFlag(adviceReq_sdPtr, ntohl(INADDR_ANY));
	if ((temp = Sock_Listen(adviceReq_sdPtr, adviceReqPort)) == SOCK_FAILURE) {
		perror("Sock_Listen()");
		exit(-1);
	}
	
	/* create and set up a socket to send advice replies to whoever */
	adviceRepl_sdPtr = Sock_Socket(SOCK_UDP);
	Sock_SetCompatible(adviceRepl_sdPtr, 1);
	Sock_SetBufferSize(adviceRepl_sdPtr, 16384);
	if (Sock_Connect(adviceRepl_sdPtr, userAddr, adviceReplPort) == SOCK_FAILURE) {
		perror("Sock_Connect()");
		exit(-1);
	}
	
	/* create and set up a socket to send/receive state messages from whoever */
/*	state_sdPtr = Sock_Socket(SOCK_UDP_MULTI);
	Sock_SetCompatible(state_sdPtr, 1);
	Sock_SetBufferSize(state_sdPtr, 16384);
	if (Sock_Connect(state_sdPtr, "224.5.6.9", statePort) == SOCK_FAILURE) {
		perror("Multicast Sock_Connect()");
		exit(-1);
	}*/
	
}

void InitIfState(RouteState *sysStatePtr, char *ifaces)
{
	int j, s;
	int af = AF_INET;
	IFace *iPtr;
	struct ifreq ifr;
	struct hostent *hp;
	Address sin;
	struct sigaction act, oact;
	int fd;
	struct ip_mobreq req;
	char *ifName, *ifAddr;
	
	/* Get list of wireless (beacon) interfaces and initialize state. */
	while ((ifName = strsep(&ifaces, " ")) != NULL){
		if (strcmp(ifName, "") == 0)	/* all interfaces done */
			break;

		if ((ifAddr = strsep(&ifaces, " "))==NULL){
			fprintf(stderr, "No address for interface %s\n", ifName);
			exit(-1);
		}
		
		/* Make sure that the interface actually exists */
		strncpy(ifr.ifr_name, ifName, sizeof(ifr.ifr_name));
		af=ifr.ifr_addr.sa_family = AF_INET;
		s = socket(af, SOCK_DGRAM, 0);
		if (s < 0) {
			perror("ifconfig: socket");
			exit(1);
		}
		
		if (ioctl(s, SIOCGIFFLAGS, (caddr_t)&ifr) < 0 ||
		    (ifr.ifr_flags & IFF_UP)==0) {
			dprintf(10, "Interface %s doesn't exist\n", ifName);
			continue;
		} else {
			dprintf(10, "Interface %s does exist\n", ifName);
		}			  
		
		iPtr = sysStatePtr->iface[sysStatePtr->numIfaces] = (IFace *) malloc(sizeof(IFace));
		if (iPtr == NULL) {
			dprintf(1, "out of memory\n");
			exit(1);
		}
		dprintf(20, "interface %s\n", ifName);
		strcpy(iPtr->ifname, ifName);
		strcpy(iPtr->locName, ifAddr);
		if ((iPtr->ifNum = GetInterface(iPtr)) == -1) {
			dprintf(1, "bad interface\n");
			exit(1);
		}
		
		/* get the netmask for the corresponding network */
		ifr.ifr_addr.sa_family = AF_INET;
		strncpy(ifr.ifr_name, iPtr->ifname, sizeof(ifr.ifr_name));
		if (ioctl(beacon_sdPtr->fd, SIOCGIFNETMASK, (caddr_t) & ifr) < 0)
			perror("ioctl(SIOCGIFNETMASK)");
		iPtr->netMask.s_addr = ((struct sockaddr_in *) & (ifr.ifr_addr))->sin_addr.s_addr;
		dprintf(20, "netmask=%x\n", iPtr->netMask.s_addr);
		
		iPtr->curBSPtr = NULL;
		iPtr->active = FALSE;
		for (j = 0; j < MAX_BS; j++) {
			iPtr->BSPtr[j] = (BSState *) malloc(sizeof(BSState));
		}
		
		/* Setup stats retrieval. Currently only for wavelan. */
		iPtr->sampdPtr = InitSample(iPtr);	/* NULL ==> non-Wavelan */
		
		/* setup interface/beacon control */
		if ((iPtr->ifs = InitBeacon(iPtr->ifname)) == -1) {
			perror("InitBeacon");
			exit(1);
		}
		
		/* setup mobile-basestation communication */
		if ((iPtr->locAddr.s_addr = inet_addr(iPtr->locName)) == -1){
			if ((hp = gethostbyname(iPtr->locName))) {
				bcopy((char *) hp->h_addr, (char *) &(sin.sin_addr),
				      sizeof(sin.sin_addr));
				iPtr->locAddr.s_addr = sin.sin_addr.s_addr;
			}
		}
		iPtr->netAddr.s_addr = iPtr->locAddr.s_addr & iPtr->netMask.s_addr;
		iPtr->sdPtr = beacon_sdPtr;	/* for beacons */
		iPtr->beaconPort = beaconPort;
		
		++sysStatePtr->numIfaces;
	}

	dprintf(10, "Done scanning for ifaces\n");
	
	if (sysStatePtr->numIfaces == 0) {
		dprintf(5, "no good interfaces specified\n");
		exit(1);
	}
	
	/* Set up the handler for the SIGALRM signal (which will happen
	   when the mobile moves out of range of an overlay) */
	act.sa_handler=AnalyzeStateWrapperSIGALRM;
	act.sa_flags=0;
	sigemptyset(&act.sa_mask);
	sigemptyset(&oact.sa_mask);
	sigaction(SIGALRM, &act, &oact);
	
	/* Register ourselves with the mobileIP code so we are
	   notified when there are many packets coming in on 1
	   interface */
	act.sa_handler=AnalyzeStateWrapperSIGUSR1;
	act.sa_flags=0;
	sigemptyset(&act.sa_mask);
	sigemptyset(&oact.sa_mask);
	sigaction(SIGUSR1, &act, &oact);

	/* Cleanup state on exit. */
	signal(SIGHUP, CleanupStateExit);
	signal(SIGINT, CleanupStateExit);
	signal(SIGQUIT, CleanupStateExit);
	signal(SIGTERM, CleanupStateExit);
	
	fd = socket(AF_INET, SOCK_STREAM, 0);
	req.burstThreshold=burstThreshold;
	req.handoffCtlrPid=getpid();
	
	j = sizeof(req);
	if (setsockopt(fd, IPPROTO_IP, IP_NOTIFY_PID,
		       (char *) &req, j) < 0)
		perror("Setting Notify pid");
	
	j = sizeof(req);
	if (setsockopt(fd, IPPROTO_IP, IP_NOTIFY_THRESHOLD,
		       (char *) &req, j) < 0)
		perror("Setting Notify threshold");
}

int main(int argc, char *argv[])
{
	RouteState sysState;
	Address sin;
	struct hostent *hp;
	char *tmpPtr;
	char ifaces[1000];
	char buf[255];
	int beaconCount=0, donesm=0;
	
	/* Some initialization stuff. Done via system() for now (XXX). */
	system("mipset_mobile -ifon -i irp");
	system("mipset_mobile -ifon -i wlp");
	system("mipset_mobile -ifon -i wl");
	system("route delete default");
	system("mipset_mobile -cm");
	
	/* Setup kernel-level stats retrieval */
	 if ((kvmd = kvm_openfiles(nlistf, memf, NULL, O_RDONLY, buf)) == NULL) {
		  fprintf(stderr, "kvm_open: %s\n", buf);
		  exit(1);
	 }
	 if (kvm_nlist(kvmd, nl) < 0 || nl[0].n_type == 0) {
		  if (nlistf)
			   fprintf(stderr, "%s: no namelist\n", nlistf);
		  else
			   fprintf(stderr, "no namelist\n");
		  exit(1);
	 }

        /* The argument ParseConfFile expects that any string pointers
	   passed as arguments have room for a strcpy() of a new value.
	   So we malloc() a new string and copy the old value to this
	   string. */
	tmpPtr=wiName;
	wiName=(char *) malloc(100*sizeof(char));
	strcpy(wiName, tmpPtr);

	tmpPtr=multiName;
	multiName=(char *) malloc(100*sizeof(char));
	strcpy(multiName, tmpPtr);

	tmpPtr=reverseTunnelAddr;
	reverseTunnelAddr=(char *) malloc(100*sizeof(char));
	strcpy(reverseTunnelAddr, tmpPtr);

        /* Use the values in the configuration file for the decap
           port, wired name, and wireless name. */
	ParseConfFile(confFileName, 7,
		      "burstThreshold", "%d", &burstThreshold,
		      "beaconPort", "%d", &beaconPort,
		      "decapPort", "%d", &decapPort,
		      "wiName", "%s", wiName,
		      "multiName", "%s", multiName,
		      "ifaces", "%[^\n]", ifaces,
		      "reverseTunnelAddr", "%s", reverseTunnelAddr);
	
	Opt_Parse(argc, argv, optionArray, Opt_Number(optionArray), 0);
	SetDebugLevel(debug);
	Sock_SigPipeHandle();
	
	memset(&sysState, 0, sizeof(sysState));

	/* Reset the list of dont touch addresses */
	ResetDontTouchAddr();
	
	/* Set up the global pointer to sysState */	 
	sysStatePtr=&sysState;

	/* Initialize interface-independent fields of sysState. */
	/* Setup route socket first. */
	if ((sysState.routes = InitRoute()) == -1) {
		perror("route socket");
		exit(1);
	}
	
	/* Setup real address and multicast address for MH. */
	if ((sysState.realAddr.s_addr = inet_addr(wiName)) == -1)
		if ((hp = gethostbyname(wiName))) {
			bcopy((char *) hp->h_addr, (char *) &(sin.sin_addr),
			      sizeof(sin.sin_addr));
			sysState.realAddr.s_addr = sin.sin_addr.s_addr;
		}
	if ((sysState.multiAddr.s_addr = inet_addr(multiName)) == -1)
		perror("bad multiaddr");

	/* Setup sockets used for communication */
	InitSockets();

	/* Initialize all state having to do with interfaces */
	InitIfState(&sysState, ifaces);

	/* Main loop: listen on all interfaces for beacon packets, add beacon
	 * if new packet arrives on some interface, delete beacons > OLD_TIME
	 * secs */
	while (1) {
		ListenForBeacons(&sysState);
		beaconCount++;
		if (beaconCount>10 && !donesm){
			dprintf(50, "Doing sm()\n");
			sprintf(buf, "mipset_mobile -sm -s %s\n", wiName);
			system(buf);
			donesm=1;
		}
	}
	
}






