/***********************************************************************
 *
 * optioneq.c
 *
 *	This is similar to option.c, but it parses options that
 *	don't start with a minus sign.  Instead, options are of
 *	the form
 *	<option>=value
 *	This allows two types of options in command line arguments.
 *	The format for input variables is the same as for Opt_Parse.
 *
 * Copyright 1993 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 ***********************************************************************
 */

#ifndef lint
static char rcsid[] = "$Header: Locker:$";
#endif /* not lint */

#include <stdio.h>
#include <string.h>
#include "option.h"

static char*	pn;	/* program name */
extern double strtosize ();
extern double strtod ();


/*----------------------------------------------------------------------
 *
 * OptNoArg --
 *
 *	Print a message complaining about the lack of an argument.
 *
 *----------------------------------------------------------------------
 */
static
void
OptNoArg (optionPtr)
Option*	optionPtr;
{
    fprintf (stderr, "Warning: %s option '-%s' needs an argument\n", pn,
	     optionPtr->key);
}

/*----------------------------------------------------------------------
 *
 * OptNonNumeric --
 *
 *----------------------------------------------------------------------
 */
static
void
OptNonNumeric (optionPtr, valStr)
Option*	optionPtr;
char*	valStr;
{
    fprintf (stderr, "Warning: '-%s' got non-numeric argument '%s'.",
	     optionPtr->key, valStr);
    fprintf (stderr, "  Setting to 0.\n");
}    

/*----------------------------------------------------------------------
 *
 * decodeArg --
 *
 *	Decode a single argument and place the appropriate value in the
 *	right variable.  Return a negative number if the conversion didn't
 *	work; otherwise, return 0 or 1 (if the next argument was used).
 *	
 *----------------------------------------------------------------------
 */
static
int
decodeArg (curOpt, valStr)
Option*	curOpt;
char*	valStr;
{
    int		numUsed = 0;
    void*	valPtr = curOpt->address;
    char*	endValStr;
    int		tmp;

    switch (curOpt->type) {
      case OPT_TRUE:
	*((int*)valPtr) = 1;
	break;
      case OPT_FALSE:
	*((int*)valPtr) = 0;
	break;
      case OPT_INT:
      case OPT_HEX:
	if (valStr == NULL) {
	    OptNoArg (curOpt);
	} else {
	    *((int*)valPtr) = strtol (valStr, &endValStr, 0);
	    if (valStr == endValStr) {
		OptNonNumeric (curOpt, valStr);
	    } else {
		numUsed = 1;
	    }
	}
	break;
      case OPT_FLOAT:
	if (valStr == NULL) {
	    OptNoArg (curOpt);
	} else {
	    *((double *)valPtr) = strtod (valStr, &endValStr, 0);
	    if (valStr == endValStr) {
		OptNonNumeric (curOpt, valStr);
	    } else {
		numUsed = 1;
	    }
	}
	break;
      case OPT_SIZE:
	if (valStr == NULL) {
	    OptNoArg (curOpt);
	} else {
	    *((int *)valPtr) = (int)strtosize (valStr, &endValStr);
	    if (valStr == endValStr) {
		OptNonNumeric (curOpt, valStr);
	    } else {
		numUsed = 1;
	    }
	}
	break;
      case OPT_SIZE_DOUBLE:
	if (valStr == NULL) {
	    OptNoArg (curOpt);
	} else {
	    *((double *)valPtr) = (double)strtosize (valStr, &endValStr);
	    if (valStr == endValStr) {
		OptNonNumeric (curOpt, valStr);
	    } else {
		numUsed = 1;
	    }
	}
	break;
      case OPT_STRING:
	if (valStr == NULL) {
	    OptNoArg (curOpt);
	} else {
	    numUsed = 1;
	    *((char**)valPtr) = valStr;
	}
	break;
      case OPT_FUNC:
	{
	    int (*handlerProc)();
	    
	    handlerProc = (int (*)())valPtr;
	    if ((*handlerProc)(curOpt->key, valStr)) {
		numUsed = 1;
	    }
	}
	break;
      case OPT_TIME:
	fprintf (stderr, "OPT_TIME not supported!\n");
	break;
      default:	/* all else is OPT_CONSTANT */
	*((int*)valPtr) = curOpt->type;
	break;
    }

    return (numUsed);
}

/*----------------------------------------------------------------------
 *
 * findOpt --
 *
 *	Match the option string to find the correct option.  If none
 *	return NULL.
 *
 *----------------------------------------------------------------------
 */
Option*
findOpt (curArg, optionArray, numOpts, flags)
char*	curArg;
Option*	optionArray;
int	numOpts;
int	flags;
{
    Option*	curOpt;
    int		i;
    extern int	strcasecmp ();

    for (i = 0, curOpt = optionArray; i < numOpts; i++, curOpt++) {
	if (!(((flags & OPT_CASE_INSENSITIVE) ? strcasecmp :
	       strcmp)(curOpt->key, curArg))) {
	    break;
	}
    }
    if (i == numOpts) {
	curOpt = NULL;
    }
    return (curOpt);
}


/*----------------------------------------------------------------------
 *
 * Opt_EqParse --
 *
 *	This routine is designed for parsing things that look like
 *
 *	foo=fooValue bar=barValue
 *
 *	It stops parsing as soon as a non-recognized option is found.
 *	This allows it to be used as a subprogram of Opt_Parse.
 *
 *----------------------------------------------------------------------
 */
int
Opt_EqParse (argc, argv, optionArray, numOpts, flags)
int	argc;
char	*argv[];
Option*	optionArray;
int	numOpts;
unsigned int flags;
{
    char	*equalSign;
    int		i;
    Option*	curOpt;
    int		argNum;
    char*	curArg;
    char*	argStr;

    flags &= ~OPT_ALLOW_CLUSTERING;	/* equals doesn't support clustering */
    
    for (argNum = 0; argNum < argc; argNum++) {
	curArg = argv[argNum];
	if ((strcmp(curArg, "?") == 0) || (strcmp(curArg, "help") == 0)) {
	    Opt_PrintUsage (argv[0], optionArray, numOpts);
	    exit(0);
	}
	equalSign = strchr (curArg, '=');
	if (equalSign != NULL) {
	    *equalSign = '\0';
	}
	argStr = equalSign + 1;
	curOpt = findOpt (curArg, optionArray, numOpts, flags);
	if (equalSign != NULL) {
	    *equalSign = '=';
	}
	if (curOpt == NULL) {
	    /*
	     * Exit loop if we didn't find the argument
	     */
	    break;
	}
	/*
	 * Always consume exactly one argument for this format.
	 */
	decodeArg (curOpt, argStr);
    }

    /*
     * Pack arguments into front of argv.
     */
    for (i = argNum; i < argc; i++) {
	argv[i - argNum] = argv[i];
    }
    return (argc - argNum);
}
